# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import xml.parsers.expat
import wx
import StringIO
import os
import tempfile
import base64
import test
from taskcoachlib import persistence, config, operating_system
from taskcoachlib.domain import date, task


class XMLTemplateReaderTestCase(test.TestCase):
    def convert(self, old_format, now=date.Now):
        return persistence.TemplateXMLReader.convert_old_format(old_format, now)
 
    def testConvertNow(self):
        self.assertEqual('now', self.convert('Now()'))

    def testConvertToday(self):
        self.assertEqual('00:00 AM today', self.convert('Today()'))

    def testConvertTomorrow(self):
        self.assertEqual('11:59 PM tomorrow', self.convert('Tomorrow()'))

    def testConvertEndOfDay(self):
        self.assertEqual('11:59 PM today', self.convert('Now().endOfDay()'))

    def testConvertTomorrow2(self):
        self.assertEqual('11:59 PM tomorrow', 
                         self.convert('Now().endOfDay() + oneDay'))

    def testConvertNowAndPositiveTimeDelta(self):
        self.assertEqual('%d minutes from now' % date.TimeDelta(2, 1861, 0).minutes(),
                         self.convert('Now() + TimeDelta(2, 1861, 0)'))

    def testConvertNowAndNegativeTimeDelta(self):
        self.assertEqual('%d minutes ago' % date.TimeDelta(2, 1861, 0).minutes(),
                         self.convert('Now() - TimeDelta(2, 1861, 0)'))

    def testConvertTodayAndPositiveTimeDelta(self):
        now = date.Now()
        expectedDate = date.Now() + date.TimeDelta(17)
        expectedDateTime = date.DateTime(expectedDate.year, expectedDate.month, 
                                         expectedDate.day)
        expectedMinutes = (expectedDateTime - now).minutes() 
        actualMinutes = int(self.convert('Today() + TimeDelta(17)').split(' ')[0])
        self.failUnless(abs(actualMinutes - expectedMinutes) <= 1)

    def testConvertTodayAndZeroTimeDelta(self):
        now = date.Now()
        expectedMinutes = (now - now.startOfDay()).minutes()
        actualMinutes = int(self.convert('Today() + TimeDelta(0)').split(' ')[0])
        self.failUnless(abs(actualMinutes - expectedMinutes) <= 1)


class XMLReaderTestCase(test.TestCase):
    tskversion = 'Subclass responsibility'

    def setUp(self):
        super(XMLReaderTestCase, self).setUp()
        task.Task.settings = config.Settings(load=False)
            
    def writeAndRead(self, xml_contents):
        # pylint: disable=W0201
        self.fd = StringIO.StringIO()
        self.fd.name = 'testfile.tsk'
        self.reader = persistence.XMLReader(self.fd)
        all_xml = '<?taskcoach release="whatever" ' \
                  'tskversion="%d"?>\n' % self.tskversion + xml_contents
        self.fd.write(all_xml)
        self.fd.seek(0)
        return self.reader.read()

    def writeAndReadTasks(self, xml_contents):
        return self.writeAndRead(xml_contents)[0]

    def writeAndReadCategories(self, xml_contents):
        return self.writeAndRead(xml_contents)[1]

    def writeAndReadNotes(self, xml_contents):
        return self.writeAndRead(xml_contents)[2]

    def writeAndReadSyncMLConfig(self, xml_contents):
        return self.writeAndRead(xml_contents)[3]

    def writeAndReadGUID(self, xml_contents):
        return self.writeAndRead(xml_contents)[5]

    def writeAndReadTasksAndCategories(self, xml_contents):
        tasks, categories, _, _, _, _ = self.writeAndRead(xml_contents)
        return tasks, categories

    def writeAndReadTasksAndCategoriesAndNotes(self, xml_contents):
        tasks, categories, notes, _, _, _ = self.writeAndRead(xml_contents)
        return tasks, categories, notes

    def writeAndReadCategoriesAndNotes(self, xml_contents):
        _, categories, notes, _, _, _ = self.writeAndRead(xml_contents)
        return categories, notes
    

class TempFileLockTest(XMLReaderTestCase):
    tskversion = 25

    def setUp(self):
        self.oldMkstemp = tempfile.mkstemp
     
        def newMkstemp(*args, **kwargs):  # pragma: no cover
            handle, name = self.oldMkstemp(*args, **kwargs)
            self.__filename = name  # pylint: disable=W0201
            return handle, name
     
        tempfile.mkstemp = newMkstemp
        super(TempFileLockTest, self).setUp()

    def tearDown(self):
        tempfile.mkstemp = self.oldMkstemp
        super(TempFileLockTest, self).tearDown()

    def testLock(self):
        if os.name == 'nt':  # pragma: no cover
            self.writeAndReadTasks(\
                '<tasks>\n<task status="0">\n'
                '<attachment type="mail" status="0">\n'
                '<data extension="eml">%s</data>\n'
                '</attachment>\n</task>\n</tasks>\n' % base64.encodestring('Data'))
            try:
                os.remove(self.__filename)
                os.remove(self.__filename + '.delta')
            except:
                pass  # pylint: disable=W0702

            self.assert_(os.path.exists(self.__filename))


class XMLReaderVersion6Test(XMLReaderTestCase):
    tskversion = 6

    def testDescription(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task description="%s" id="foo"/>
        </tasks>\n''' % u'Description')
        self.assertEqual(u'Description', tasks[0].description())

    def testEffortDescription(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="foo">
                <effort start="2004-01-01 10:00:00.123000" 
                        stop="2004-01-01 10:30:00.123000" 
                        description="Yo"/>
            </task>
        </tasks>''')
        self.assertEqual('Yo', tasks[0].efforts()[0].description())
     

class XMLReaderVersion8Test(XMLReaderTestCase):   
    tskversion = 8

    def testReadTaskWithoutPriority(self):
        tasks = self.writeAndReadTasks('<tasks><task id="foo"/></tasks>')
        self.assertEqual(0, tasks[0].priority())
     
     
class XMLReaderVersion9Test(XMLReaderTestCase):
    tskversion = 9
 
    def testReadTaskWithoutId(self):
        tasks = self.writeAndReadTasks('<tasks><task id="foo"/></tasks>')
        self.failUnless(tasks[0].id())
     

class XMLReaderVersion10Test(XMLReaderTestCase):
    tskversion = 10
 
    def testReadTaskWithoutFee(self):
        tasks = self.writeAndReadTasks('<tasks><task id="foo"/></tasks>')
        self.assertEqual(0, tasks[0].hourlyFee())
        self.assertEqual(0, tasks[0].fixedFee())


class XMLReaderVersion11Test(XMLReaderTestCase):
    tskversion = 11
 
    def testReadTaskWithoutReminder(self):
        tasks = self.writeAndReadTasks('<tasks><task id="foo"/></tasks>')
        self.assertEqual(date.DateTime(), tasks[0].reminder())
     
     
class XMLReaderVersion12Test(XMLReaderTestCase):
    tskversion = 12
 
    def testReadTaskWithoutMarkCompletedWhenAllChildrenCompletedSetting(self):
        tasks = self.writeAndReadTasks('<tasks><task id="foo"/></tasks>')
        self.assertEqual(None, 
                         tasks[0].shouldMarkCompletedWhenAllChildrenCompleted())
     

class XMLReaderVersion13Test(XMLReaderTestCase):
    tskversion = 13

    def testOneCategory(self):
        tasks, categories = self.writeAndReadTasksAndCategories('''
        <tasks>
            <task id="1">
                <category>test</category>
            </task>
        </tasks>''')

        self.assertEqual('test', categories[0].subject())
        self.assertEqual(set([tasks[0]]), categories[0].categorizables())
        self.assertEqual(set([categories[0]]), tasks[0].categories())

    def testMultipleCategories(self):
        tasks, categories = self.writeAndReadTasksAndCategories('''
        <tasks>
            <task id="1">
                <category>test</category>
                <category>another</category>
                <category>yetanother</category>
            </task>
        </tasks>''')

        for category in categories:
            self.assertEqual(set([tasks[0]]), category.categorizables())
            self.failUnless(category in tasks[0].categories())
         
    def testSubTaskWithCategories(self):
        tasks, categories = self.writeAndReadTasksAndCategories('''
        <tasks>
            <task id="1">
                <category>test</category>
                <task id="1.1">
                    <category>another</category>
                </task>
            </task>
        </tasks>''')
        testCategory = categories[0]
        anotherCategory = categories[1]
        self.assertEqual("1", list(testCategory.categorizables())[0].id())
        self.assertEqual("1.1", list(anotherCategory.categorizables())[0].id())         
        self.assertEqual(set([testCategory]), tasks[0].categories())
        self.assertEqual(set([anotherCategory]), tasks[0].children()[0].categories())
     
     
class XMLReaderVersion14Test(XMLReaderTestCase):
    tskversion = 14
 
    def testEffortWithMilliseconds(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <effort start="2004-01-01 10:00:00.123000" 
                        stop="2004-01-01 10:30:00.123000"/>
            </task>
        </tasks>''')
        self.assertEqual(1, len(tasks[0].efforts()))
        self.assertEqual(date.TimeDelta(minutes=30), tasks[0].timeSpent())
        self.assertEqual(tasks[0], tasks[0].efforts()[0].task())


class XMLReaderVersion16Text(XMLReaderTestCase):
    tskversion = 16

    def testOneAttachmentCompat(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <attachment>whatever.tsk</attachment>
            </task>
        </tasks>''')
        self.assertEqual(['whatever.tsk'], 
                         [att.location() for att in tasks[0].attachments()])
        self.assertEqual(['whatever.tsk'], 
                         [att.subject() for att in tasks[0].attachments()])
     
    def testTwoAttachmentsCompat(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <attachment>whatever.tsk</attachment>
                <attachment>another.txt</attachment>
            </task>
        </tasks>''')
        self.assertEqual(['whatever.tsk', 'another.txt'], 
                         [att.location() for att in tasks[0].attachments()])
        self.assertEqual(['whatever.tsk', 'another.txt'], 
                         [att.subject() for att in tasks[0].attachments()])
     
    def testOneAttachment(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <attachment>FILE:whatever.tsk</attachment>
            </task>
        </tasks>''')
        self.assertEqual(['whatever.tsk'], 
                         [att.location() for att in tasks[0].attachments()])
        self.assertEqual(['whatever.tsk'], 
                         [att.subject() for att in tasks[0].attachments()])
     
    def testTwoAttachments(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <attachment>FILE:whatever.tsk</attachment>
                <attachment>FILE:another.txt</attachment>
            </task>
        </tasks>''')
        self.assertEqual(['whatever.tsk', 'another.txt'], 
                         [att.location() for att in tasks[0].attachments()])
        self.assertEqual(['whatever.tsk', 'another.txt'], 
                         [att.subject() for att in tasks[0].attachments()])


# There's no XMLReaderVersion17Test because the only difference between version
# 17 and 18 is the addition of an optional color attribute to categories in
# version 18. So the tests for version 18 test version 17 as well.


class XMLReaderVersion18Test(XMLReaderTestCase):
    tskversion = 18

    def testLastModificationTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task lastModificationTime="2004-01-01 10:00:00"/>
        </tasks>''')
        self.assertEqual(1, len(tasks))  # Ignore lastModificationTime


class XMLReaderVersion19Test(XMLReaderTestCase):
    tskversion = 19  # New in release 0.69.0?        

    def testDailyRecurrence(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task recurrence="daily"/>
        </tasks>''')
        self.assertEqual('daily', tasks[0].recurrence().unit)    

    def testWeeklyRecurrence(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task recurrence="weekly"/>
        </tasks>''')
        self.assertEqual('weekly', tasks[0].recurrence().unit)

    def testMonthlyRecurrence(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task recurrence="monthly"/>
        </tasks>''')
        self.assertEqual('monthly', tasks[0].recurrence().unit)
             
    def testRecurrenceCount(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task recurrenceCount="10"/>
        </tasks>''')
        self.assertEqual(10, tasks[0].recurrence().count)
     
    def testMaxRecurrenceCount(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task maxRecurrenceCount="10"/>
        </tasks>''')
        self.assertEqual(10, tasks[0].recurrence().max)

    def testRecurrenceFrequency(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task recurrenceFrequency="3"/>
        </tasks>''')
        self.assertEqual(3, tasks[0].recurrence().amount)


class XMLReaderVersion20Test(XMLReaderTestCase):
    tskversion = 20  # New in release 0.71.0
        
    def testReadEmptyStream(self):
        reader = persistence.XMLReader(StringIO.StringIO())
        try:
            reader.read()
            self.fail('Expected ExpatError or ParseError')  # pragma: no cover
        except xml.parsers.expat.ExpatError:
            pass  # pragma: no cover
        except xml.etree.ElementTree.ParseError:
            pass  # pragma: no cover
     
    def testNoTasksAndNoCategories(self):
        tasks, categories, notes = self.writeAndReadTasksAndCategoriesAndNotes('<tasks/>\n')
        self.assertEqual(([], [], []), (tasks, categories, notes))
     
    def testOneTask(self):
        tasks = self.writeAndReadTasks('<tasks><task/></tasks>\n')
        self.assertEqual(1, len(tasks))
        self.assertEqual('', tasks[0].subject())

    def testTwoTasks(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task subject="1"/>
            <task subject="2"/>
        </tasks>\n''')
        self.assertEqual(2, len(tasks))
        self.assertEqual('1', tasks[0].subject())
        self.assertEqual('2', tasks[1].subject())
     
    def testOneTask_Subject(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task subject="Yo"/>
        </tasks>\n''')
        self.assertEqual('Yo', tasks[0].subject())
     
    def testOneTask_UnicodeSubject(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task subject="???"/>
        </tasks>\n''')
        self.assertEqual('???', tasks[0].subject())        
     
    def testBudget(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task budget="4:10:10"/>
        </tasks>\n''')
        self.assertEqual(date.TimeDelta(hours=4, minutes=10, seconds=10), 
                         tasks[0].budget())

    def testBudget_NoBudget(self):
        tasks = self.writeAndReadTasks('<tasks><task/></tasks>\n')
        self.assertEqual(date.TimeDelta(), tasks[0].budget())
     
    def testDescription(self):
        description = u'Description\nline 2'
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <description>%s</description>
            </task>
        </tasks>\n''' % description)
        self.assertEqual(description, tasks[0].description())
 
    def testNoChildren(self):
        tasks = self.writeAndReadTasks('<tasks><task/></tasks>\n')
        self.failIf((tasks[0].children()))
     
    def testChild(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <task/>
            </task>
        </tasks>\n''')
        self.assertEqual(1, len(tasks[0].children()))
        self.assertEqual(1, len(tasks))
     
    def testChildren(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <task/>
                <task/>
            </task>
        </tasks>\n''')
        self.assertEqual(2, len(tasks[0].children()))
        self.assertEqual(1, len(tasks))
     
    def testGrandchild(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <task>
                    <task/>
                </task>
            </task>
        </tasks>\n''')
        self.assertEqual(1, len(tasks))
        parent = tasks[0]
        self.assertEqual(1, len(parent.children()))
        self.assertEqual(1, len(parent.children()[0].children()))
     
    def testEffort(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <effort start="2004-01-01 10:00:00" 
                        stop="2004-01-01 10:30:00"/>
            </task>
        </tasks>''')
        self.assertEqual(1, len(tasks[0].efforts()))
        self.assertEqual(date.TimeDelta(minutes=30), tasks[0].timeSpent())
        self.assertEqual(tasks[0], tasks[0].efforts()[0].task())
 
    def testChildEffort(self):    
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <task>
                    <effort start="2004-01-01 10:00:00" 
                            stop="2004-01-01 10:30:00"/>
                </task>
            </task>
        </tasks>''')
        child = tasks[0].children()[0]
        self.assertEqual(1, len(child.efforts()))
        self.assertEqual(date.TimeDelta(minutes=30), child.timeSpent())
        self.assertEqual(child, child.efforts()[0].task())

    def testEffortDescription(self):
        description = u'Description\nLine 2'
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <effort start="2004-01-01 10:00:00">
                    <description>%s</description>
                </effort>
            </task>
        </tasks>''' % description)
        self.assertEqual(description, tasks[0].efforts()[0].description())
     
    def testActiveEffort(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <effort start="2004-01-01 10:00:00"/>
            </task>
        </tasks>''')
        self.assertEqual(1, len(tasks[0].efforts()))
        self.failUnless(tasks[0].isBeingTracked())
             
    def testPriority(self):
        tasks = self.writeAndReadTasks('<tasks><task priority="5"/></tasks>')        
        self.assertEqual(5, tasks[0].priority())
     
    def testTaskId(self):
        tasks = self.writeAndReadTasks('<tasks><task id="xyz"/></tasks>')
        self.assertEqual('xyz', tasks[0].id())

    def testTaskColor(self):        
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task color="(255, 0, 0, 255)"/>
        </tasks>''')
        self.assertEqual(wx.RED, tasks[0].backgroundColor())
             
    def testHourlyFee(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task hourlyFee="100"/>
            <task hourlyFee="5.5"/>
        </tasks>''')
        self.assertEqual(100, tasks[0].hourlyFee())
        self.assertEqual(5.5, tasks[1].hourlyFee())
     
    def testFixedFee(self):
        tasks = self.writeAndReadTasks('<tasks><task fixedFee="240.50"/></tasks>')
        self.assertEqual(240.5, tasks[0].fixedFee())

    def testNoReminder(self):
        tasks = self.writeAndReadTasks('<tasks><task reminder="None"/></tasks>')
        self.assertEqual(date.DateTime(), tasks[0].reminder())
     
    def testReminder(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task reminder="2004-01-01 10:00:00"/>
        </tasks>''')
        self.assertEqual(date.DateTime(2004, 1, 1, 10, 0, 0, 0),
                         tasks[0].reminder())
     
    def testMarkCompletedWhenAllChildrenCompletedSetting_True(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task shouldMarkCompletedWhenAllChildrenCompleted="True"/>
        </tasks>''')
        self.assertEqual(True, 
                         tasks[0].shouldMarkCompletedWhenAllChildrenCompleted())

    def testMarkCompletedWhenAllChildrenCompletedSetting_False(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task shouldMarkCompletedWhenAllChildrenCompleted="False"/>
        </tasks>''')
        self.assertEqual(False,
                         tasks[0].shouldMarkCompletedWhenAllChildrenCompleted())

    def testMarkCompletedWhenAllChildrenCompletedSetting_None(self):
        tasks = self.writeAndReadTasks('<tasks><task/></tasks>')
        self.assertEqual(None,
                         tasks[0].shouldMarkCompletedWhenAllChildrenCompleted())

    def testTaskWithoutAttachments(self):
        tasks = self.writeAndReadTasks('<tasks><task/></tasks>')
        self.assertEqual([], tasks[0].attachments())

    def testNoteWithoutAttachments(self):
        notes = self.writeAndReadNotes('<tasks><note/></tasks>')
        self.assertEqual([], notes[0].attachments())

    def testCategoryWithoutAttachments(self):
        categories = self.writeAndReadCategories('<tasks><category/></tasks>')
        self.assertEqual([], categories[0].attachments())
     
    def testTaskWithOneAttachment(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <attachment type="file">
                    <description>whatever.tsk</description>
                    <data>whateverdata.tsk</data>
                </attachment>
            </task>
        </tasks>''')
        self.assertEqual([os.path.join(os.getcwd(),
                                       'testfile_attachments',
                                       'whateverdata.tsk')],
                         [att.location() for att in tasks[0].attachments()])
        self.assertEqual(['whatever.tsk'], [att.subject() for att in tasks[0].attachments()])

    def testNoteWithOneAttachment(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <note>
                <attachment type="file">
                    <description>whatever.tsk</description>
                    <data>whateverdata.tsk</data>
                </attachment>
            </note>
        </tasks>''')
        self.assertEqual([os.path.join(os.getcwd(),
                                       'testfile_attachments',
                                       'whateverdata.tsk')],
                         [att.location() for att in notes[0].attachments()])
        self.assertEqual(['whatever.tsk'], [att.subject() for att in notes[0].attachments()])

    def testCategoryWithOneAttachment(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category>
                <attachment type="file">
                    <description>whatever.tsk</description>
                    <data>whateverdata.tsk</data>
                </attachment>
            </category>
        </tasks>''')
        self.assertEqual([os.path.join(os.getcwd(),
                                       'testfile_attachments',
                                       'whateverdata.tsk')],
                         [att.location() for att in categories[0].attachments()])
        self.assertEqual(['whatever.tsk'], 
                         [att.subject() for att in categories[0].attachments()])
     
    def testTaskWithTwoAttachments(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <attachment type="file">
                   <description>whatever.tsk</description>
                   <data>whateverdata.tsk</data>
                </attachment>
                <attachment type="file">
                    <description>another.txt</description>
                    <data>anotherdata.txt</data>
                </attachment>
            </task>
        </tasks>''')
        self.assertEqual([os.path.join(os.getcwd(),
                                       'testfile_attachments',
                                       'whateverdata.tsk'),
                          os.path.join(os.getcwd(),
                                       'testfile_attachments',
                                       'anotherdata.txt')],
                         [att.location() for att in tasks[0].attachments()])
        self.assertEqual(['whatever.tsk', 'another.txt'], 
                         [att.subject() for att in tasks[0].attachments()])
     
    def testOneCategory(self):
        categories = self.writeAndReadCategories('<tasks><category subject="cat"/></tasks>')
        self.assertEqual('cat', categories[0].subject())

    def testTwoCategories(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category subject="cat1"/>
            <category subject="cat2"/>
        </tasks>''')
        self.assertEqual(['cat1', 'cat2'], 
            [category.subject() for category in categories])

    def testCategoryId(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category id="catId"/>
        </tasks>''')
        self.assertEqual('catId', categories[0].id())
         
    def testCategoryWithDescription(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category subject="cat">
                <description>Description</description>
            </category>
        </tasks>''')
        self.assertEqual('Description', categories[0].description())
 
    def testCategoryColor(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category subject="cat" color="(255, 0, 0, 255)"/>
        </tasks>''')
        self.assertEqual(wx.RED, categories[0].backgroundColor())
     
    def testOneTaskWithCategory(self):
        tasks, categories = self.writeAndReadTasksAndCategories('''
        <tasks>
            <category subject="cat" categorizables="1"/>
            <task id="1"/>
        </tasks>''')
        self.assertEqual(set(tasks), categories[0].categorizables())
     
    def testTwoRecursiveCategories(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category subject="cat1">
                <category subject="cat1.1"/>
            </category>
        </tasks>''')
        self.assertEqual('cat1.1', categories[0].children()[0].subject())
     
    def testRecursiveCategoriesNotInResultList(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category subject="cat1">
                <category subject="cat1.1"/>
            </category>
        </tasks>''')
        self.assertEqual(1, len(categories))

    def testRecursiveCategoriesWithTwoTasks(self):
        tasks, categories = self.writeAndReadTasksAndCategories('''
        <tasks>
            <category subject="cat1" categorizables="1">
                <category subject="cat1.1" categorizables="2"/>
            </category>
            <task subject="task1" id="1"/>
            <task subject="task2" id="2"/>
        </tasks>''')

        self.assertEqual(tasks[0], list(categories[0].categorizables())[0])
        self.assertEqual(tasks[1], 
                         list(categories[0].children()[0].categorizables())[0])
     
    def testSubtaskCategory(self):
        tasks, categories = self.writeAndReadTasksAndCategories('''
        <tasks>
            <category subject="cat1" categorizables="1.1"/>
            <task subject="task1" id="1">
                <task subject="task2" id="1.1"/>
            </task>
        </tasks>''')
        self.assertEqual(tasks[0].children()[0], 
                         list(categories[0].categorizables())[0])
     
    def testFilteredCategory(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category filtered="True" subject="category"/>
        </tasks>''')
        self.failUnless(categories[0].isFiltered())
 
    def testCategoryWithDeletedTasks(self):
        ''' There's a bug in release 0.61.5 that causes the task file to contain
            references to deleted tasks. Ignore these when loading the task 
            file.'''
        categories = self.writeAndReadCategories('''
        <tasks>
            <category subject="cat" tasks="some_task_id"/>
        </tasks>''')
        self.failIf(categories[0].categorizables())
     
    def testNote(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <note/>
        </tasks>''')
        self.failUnless(notes)

    def testNoteSubject(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <note subject="Note"/>
        </tasks>''')
        self.assertEqual('Note', notes[0].subject())

    def testNoteDescription(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <note>
                <description>Description</description>
            </note>
        </tasks>''')
        self.assertEqual('Description', notes[0].description())

    def testNoteChild(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <note>
                <note/>
            </note>
        </tasks>''')
        self.assertEqual(1, len(notes[0].children()))

    def testNoteChildWithAttachment(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <note>
                <note>
                    <attachment type="file">
                       <description>whatever.tsk</description>
                       <data>whateverdata.tsk</data>
                    </attachment>
                </note>
            </note>
        </tasks>''')
        self.assertEqual([os.path.join(os.getcwd(),
                                       'testfile_attachments',
                                       'whateverdata.tsk')],
                         [att.location() for att in notes[0].children()[0].attachments()])
        self.assertEqual(['whatever.tsk'], 
                         [att.subject() for att in notes[0].children()[0].attachments()])
     
    def testNoteCategory(self):
        categories, notes = self.writeAndReadCategoriesAndNotes('''
        <tasks>
            <note id="noteId" subject="Note"/>
            <category categorizables="noteId" subject="Category"/>
        </tasks>''')
        self.assertEqual(notes[0], list(categories[0].categorizables())[0])
     
    def testNoteId(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <note id="noteId"/>
        </tasks>''')
        self.assertEqual('noteId', notes[0].id())
     
    def testNoteColor(self):        
        notes = self.writeAndReadNotes('''
        <tasks>
            <note color="(255, 0, 0, 255)"/>
        </tasks>''')
        self.assertEqual(wx.RED, notes[0].backgroundColor())
     
    def testNoRecurrence(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task/>
        </tasks>''')
        self.failIf(tasks[0].recurrence())

    def testDailyRecurrence(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task><recurrence unit="daily"/></task>
        </tasks>''')
        self.assertEqual('daily', tasks[0].recurrence().unit)    

    def testWeeklyRecurrence(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task><recurrence unit="weekly"/></task>
        </tasks>''')
        self.assertEqual('weekly', tasks[0].recurrence().unit)    

    def testRecurrenceAmount(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task><recurrence unit="daily" amount="2"/></task>
        </tasks>''')
        self.assertEqual(2, tasks[0].recurrence().amount)    

    def testRecurrenceMax(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task><recurrence unit="daily" max="2"/></task>
        </tasks>''')
        self.assertEqual(2, tasks[0].recurrence().max)    

    def testRecurrenceCount(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task><recurrence unit="daily" count="2"/></task>
        </tasks>''')
        self.assertEqual(2, tasks[0].recurrence().count)    

    def testRecurrenceSameWeekday(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task><recurrence unit="daily" sameWeekday="True"/></task>
        </tasks>''')
        self.failUnless(tasks[0].recurrence().sameWeekday)
     
    def testTaskWithNote(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <note/>
            </task> 
        </tasks>''')
        self.assertEqual(1, len(tasks[0].notes()))
     
    def testTaskWithNotes(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <note/><note/>
            </task> 
        </tasks>''')
        self.assertEqual(2, len(tasks[0].notes()))
     
    def testTaskWithNestedNotes(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <note>
                    <note/>
                </note>
            </task> 
        </tasks>''')
        self.assertEqual(1, len(tasks[0].notes()[0].children()))
     
    def testTaskNotesDontGetAddedToOverallNotesList(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <task>
                <note/>
            </task> 
        </tasks>''')
        self.failIf(notes)
     
    def testCategoryWithNote(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category>
                <note/>
            </category> 
        </tasks>''')
        self.assertEqual(1, len(categories[0].notes()))

    def testCategoryWithNotes(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category>
                <note/><note/>
            </category> 
        </tasks>''')
        self.assertEqual(2, len(categories[0].notes()))

    def testCategoryWithNestedNotes(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category>
                <note>
                    <note/>
                </note>
            </category> 
        </tasks>''')
        self.assertEqual(1, len(categories[0].notes()[0].children()))

    def testCategoryNotesDontGetAddedToOverallNotesList(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <category>
                <note/>
            </category> 
        </tasks>''')
        self.failIf(notes)

    def testTaskExpansion(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task expandedContexts="('None',)"/>
        </tasks>''')
        self.failUnless(tasks[0].isExpanded())

    def testTaskExpansion_MultipleContexts(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task expandedContexts="('None','Test')"/>
        </tasks>''')
        self.failUnless(tasks[0].isExpanded(context='Test'))

    def testCategoryExpansion(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category expandedContexts="('None',)"/>
        </tasks>''')
        self.failUnless(categories[0].isExpanded())

    def testNoteExpansion(self):
        notes = self.writeAndReadNotes('''
        <tasks>
            <note expandedContexts="('None',)"/>
        </tasks>''')
        self.failUnless(notes[0].isExpanded())


class XMLReaderVersion21Test(XMLReaderTestCase):
    tskversion = 21  # New in release 0.71.0

    def testAttachmentLocation(self):
        categories = self.writeAndReadCategories('''
        <tasks>
            <category>
                <attachment type="file" location="location">
                   <description>description</description>
                </attachment>
            </category>
        </tasks>''')
        self.assertEqual(['location'], [os.path.split(att.location())[-1] for att in categories[0].attachments()])


class XMLReaderVersion22Test(XMLReaderTestCase):
    tskversion = 22

    def testStatus(self):
        tasks = self.writeAndReadTasks(\
            '<tasks><task subject="Task" status="2"></task></tasks>')
        self.assertEqual(2, tasks[0].getStatus())


class XMLReaderVersion23Test(XMLReaderTestCase):
    tskversion = 23

    def testDescription(self):
        tasks = self.writeAndReadTasks(\
            '<tasks><task subject="Task" status="0">'
            '<description>\nDescription\n</description>'
            '</task></tasks>')
        self.assertEqual('\nDescription\n', tasks[0].description())

    def testAttachmentData(self):
        tasks = self.writeAndReadTasks(\
            '<tasks>\n<task status="0">\n'
            '<attachment type="mail" status="0">\n'
            '<data extension="eml">%s</data>\n'
            '</attachment>\n</task>\n</tasks>\n' % base64.encodestring('Data'))
        self.assertEqual('Data', tasks[0].attachments()[0].data())
 
    def testGUID(self):
        guid = self.writeAndReadGUID('<tasks><guid>GUID</guid></tasks>')
        self.assertEqual('GUID', guid)

    def testSyncMLConfig(self):
        syncmlConfig = self.writeAndReadSyncMLConfig('<tasks><syncml><property name="name">value</property></syncml></tasks>')
        self.assertEqual('value', syncmlConfig.get('name'))
     
     
class XMLReaderVersion24Test(XMLReaderTestCase):
    tskversion = 24  # New in release 0.72.9

    # tskversion 24 introduces newlines so that the XML is not on one long
    # line anymore. We have to be sure not to introduce new lines in
    # text nodes though.

    def testDescription(self):
        tasks = self.writeAndReadTasks(\
            '<tasks>\n<task subject="Task" status="0">\n'
            '<description>\nDescription\n</description>\n'
            '</task>\n</tasks>\n')
        self.assertEqual('Description', tasks[0].description())
     
    def testAttachmentData(self):
        tasks = self.writeAndReadTasks(\
            '<tasks>\n<task status="0">\n'
            '<attachment type="mail" status="0">\n'
            '<data extension="eml">\n%s\n</data>\n'
            '</attachment>\n</task>\n</tasks>\n' % base64.encodestring('Data'))
        self.assertEqual('Data', tasks[0].attachments()[0].data())

    def testGUID(self):
        guid = self.writeAndReadGUID('<tasks>\n<guid>\nGUID\n</guid>\n</tasks>')
        self.assertEqual('GUID', guid)

    def testSyncMLConfig(self):
        syncmlConfig = self.writeAndReadSyncMLConfig(\
            '<tasks>\n<syncml>\n'
            '<property name="name">\nvalue\n</property>\n'
            '</syncml>\n</tasks>\n')
        self.assertEqual('value', syncmlConfig.get('name'))
     
    def testSyncMLConfigWithNewLinesInXMLNodes(self):
        ''' Release 0.72.9 (and earlier?) had a bug where tags in the syncml
            config information would be split across multiple lines. Fixed in
            release 0.72.10. ''' 
        expectedGUID = '0000011d209a4b6c3f9f7c32000a00b100240032'
        actualGUID = self.writeAndReadGUID(\
            '<tasks>\n<syncml><TaskCoach-\n'
            '0000011d209a4b6c3f9f7c32000a00b100240032\n'
            '><spds><sources><TaskCoach-\n'
            '0000011d209a4b6c3f9f7c32000a00b100240032\n'
            '.Tasks/><TaskCoach-\n'
            '0000011d209a4b6c3f9f7c32000a00b100240032\n'
            '.Notes/></sources><syncml><Auth/><Conn/></syncml></spds></TaskCoach-\n'
            '0000011d209a4b6c3f9f7c32000a00b100240032\n'
            '><TaskCoach-0000011d209a4b6c3f9f7c32000a00b100240032><spds><sources><TaskCoach-0000011d209a4b6c3f9f7c32000a00b100240032.Tasks/><TaskCoach-0000011d209a4b6c3f9f7c32000a00b100240032.Notes/></sources><syncml><Auth/><Conn/></syncml></spds></TaskCoach-0000011d209a4b6c3f9f7c32000a00b100240032></syncml><guid>\n'
            '0000011d209a4b6c3f9f7c32000a00b100240032\n'
            '</guid></tasks>')
        self.assertEqual(expectedGUID, actualGUID)
     

class XMLReaderVersion26Test(XMLReaderTestCase):
    tskversion = 26  # New in release 0.75.0

    # Release 0.75.0 introduces percentage complete for tasks
 
    def testPercentageComplete(self):   
        tasks = self.writeAndReadTasks(\
            '<tasks>\n<task subject="Task" status="0" percentageComplete="50"/>\n'
            '</tasks>\n')
        self.assertEqual(50, tasks[0].percentageComplete())


class XMLReaderVersion27Test(XMLReaderTestCase):
    tskversion = 27  # New in release 0.76.0
 
    # Release 0.76.0 introduces exclusive subcategories
 
    def testExclusiveSubcategories(self):
        categories = self.writeAndReadCategories(\
            '<categories>\n'
            '<category subject="Category" exclusiveSubcategories="True"'
            ' status="0"/>\n'
            '</categories>\n')
        self.failUnless(categories[0].hasExclusiveSubcategories())
     
    def testNoExclusiveSubcategoriesByDefault(self):
        categories = self.writeAndReadCategories(\
            '<categories>\n'
            '<category subject="Category" status="0"/>\n'
            '</categories>\n')
        self.failIf(categories[0].hasExclusiveSubcategories())
     

class XMLReaderVersion28Test(XMLReaderTestCase):
    tskversion = 28  # New in release 0.78.0
 
    # Release 0.78.0 introduces foreground colors and fonts that can be set per object.

    def testTaskForegroundColor(self):
        tasks = self.writeAndReadTasks(\
            '<tasks>\n<task subject="Task" fgColor="(255,0,0)"/>\n'
            '</tasks>\n')
        self.assertEqual(wx.RED, tasks[0].foregroundColor())
 
    def testTaskBackgroundColor(self):
        tasks = self.writeAndReadTasks(\
            '<tasks>\n<task subject="Task" bgColor="(255,0,0)"/>\n'
            '</tasks>\n')
        self.assertEqual(wx.RED, tasks[0].backgroundColor())

    def testCategoryForegroundColor(self):
        categories = self.writeAndReadCategories(\
            '<categories>\n<category subject="Category" fgColor="(255,0,0)"/>\n'
            '</categories>\n')
        self.assertEqual(wx.RED, categories[0].foregroundColor())
     
    def testCategoryBackgroundColor(self):
        categories = self.writeAndReadCategories(\
            '<categories>\n<category subject="Category" bgColor="(255,0,0)"/>\n'
            '</categories>\n')
        self.assertEqual(wx.RED, categories[0].backgroundColor())
     
    def testNoteForegroundColor(self):
        notes = self.writeAndReadNotes(\
            '<notes>\n<note subject="Note" fgColor="(255,0,0)"/>\n'
            '</notes>\n')
        self.assertEqual(wx.RED, notes[0].foregroundColor())
     
    def testNoteBackgroundColor(self):
        notes = self.writeAndReadNotes(\
            '<notes>\n<note subject="Note" bgColor="(255,0,0)"/>\n'
            '</notes>\n')
        self.assertEqual(wx.RED, notes[0].backgroundColor())
     
    def testTaskFont(self):
        tasks = self.writeAndReadTasks(\
            '<tasks>\n<task subject="Task" font="%s"/>\n'
            '</tasks>\n' % wx.NORMAL_FONT.GetNativeFontInfoDesc())
        self.assertEqual(wx.NORMAL_FONT, tasks[0].font())
     
    def testNoTaskFont(self):
        tasks = self.writeAndReadTasks(\
            '<tasks>\n<task subject="Task"/>\n</tasks>\n')
        self.assertEqual(None, tasks[0].font())

    def testCategoryFont(self):
        categories = self.writeAndReadCategories(\
            '<categories>\n<category subject="Category" font="%s"/>\n'
            '</categories>\n' % wx.NORMAL_FONT.GetNativeFontInfoDesc())
        self.assertEqual(wx.NORMAL_FONT, categories[0].font())

    def testNoteFont(self):
        notes = self.writeAndReadNotes(\
            '<notes>\n<note subject="Note" font="%s"/>\n'
            '</notes>\n' % wx.NORMAL_FONT.GetNativeFontInfoDesc())
        self.assertEqual(wx.NORMAL_FONT, notes[0].font())

    def testAttachmentFont(self):
        tasks = self.writeAndReadTasks(\
            '<tasks>\n<task subject="Task">\n'
            '<attachment type="file" location="whatever" font="%s"/>\n'
            '</task>\n</tasks>\n' % wx.NORMAL_FONT.GetNativeFontInfoDesc())
        self.assertEqual(wx.NORMAL_FONT, tasks[0].attachments()[0].font())


class XMLReaderVersion29Test(XMLReaderTestCase):
    tskversion = 29

    def testEffortNewId(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="foo">
                <effort id="foobar" start="2004-01-01 10:00:00.123000" 
                        stop="2004-01-01 10:30:00.123000"
                        status="1"
                        description="Yo"/>
            </task>
        </tasks>''')
        self.assertEqual('foobar', tasks[0].efforts()[0].id())

    def testTaskIcon(self):
        tasks = self.writeAndReadTasks('<tasks><task icon="icon"/></tasks>')
        self.assertEqual('icon', tasks[0].icon())

    def testSelectedTaskIcon(self):
        tasks = self.writeAndReadTasks('<tasks><task selectedIcon="icon"/></tasks>')
        self.assertEqual('icon', tasks[0].selectedIcon())

    def testNoteIcon(self):
        notes = self.writeAndReadNotes('<tasks><note icon="icon"/></tasks>')
        self.assertEqual('icon', notes[0].icon())

    def testSelectedNoteIcon(self):
        notes = self.writeAndReadNotes('<tasks><note selectedIcon="icon"/></tasks>')
        self.assertEqual('icon', notes[0].selectedIcon())

    def testCategoryIcon(self):
        categories = self.writeAndReadCategories('<tasks><category icon="icon"/></tasks>')
        self.assertEqual('icon', categories[0].icon())

    def testSelectedCategoryIcon(self):
        categories = self.writeAndReadCategories('<tasks><category selectedIcon="icon"/></tasks>')
        self.assertEqual('icon', categories[0].selectedIcon())

    def testAttachmentIcon(self):
        tasks = self.writeAndReadTasks(\
            '<tasks><task subject="Task">'
            '<attachment type="file" location="whatever" icon="icon"/>'
            '</task></tasks>')
        self.assertEqual('icon', tasks[0].attachments()[0].icon())

    def testSelectedAttachmentIcon(self):
        tasks = self.writeAndReadTasks(\
            '<tasks><task subject="Task">'
            '<attachment type="file" location="whatever" selectedIcon="icon"/>'
            '</task></tasks>')
        self.assertEqual('icon', tasks[0].attachments()[0].selectedIcon())


class XMLReaderVersion30Test(XMLReaderTestCase):
    tskversion = 30  # New in release 1.1.0.
 
    def testStartDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task startdate="2005-04-17 10:05:11"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 4, 17, 10, 5, 11),
                         tasks[0].plannedStartDateTime())

    def testStartDateTimeWithoutTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task startdate="2005-04-17"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 4, 17), 
                         tasks[0].plannedStartDateTime())

    def testNoStartDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task />
        </tasks>\n''')
        self.assertEqual(date.DateTime(), tasks[0].plannedStartDateTime())

    def testStartDateTimeWithMicroseconds(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task startdate="2005-01-01 22:01:30.456"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 1, 1, 22, 1, 30, 456),
                         tasks[0].plannedStartDateTime())
     
    def testDueDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task duedate="2005-04-17 13:05:59"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 4, 17, 13, 5, 59),
                         tasks[0].dueDateTime())

    def testDueDateTimeWithoutTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task duedate="2005-04-17"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 4, 17, 23, 59, 59, 999999),
                         tasks[0].dueDateTime())

    def testDueDateTimeWithoutTimeWhenEndHourIs24(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task duedate="2005-04-17"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 4, 17, 23, 59, 59, 999999),
                         tasks[0].dueDateTime())

    def testNoDueDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task />
        </tasks>\n''')
        self.assertEqual(date.DateTime(), tasks[0].dueDateTime())

    def testDueDateTimeWithMicroseconds(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task duedate="2005-01-01 22:01:30.456000"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 1, 1, 22, 1, 30, 456000),
                         tasks[0].dueDateTime())
     
    def testCompletionDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task completiondate="2005-01-01 22:01:30"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 1, 1, 22, 1, 30),
                         tasks[0].completionDateTime())
        self.failUnless(tasks[0].completed())

    def testCompletionDateTimeWithMicroseconds(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task completiondate="2005-01-01 22:01:30.456000"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 1, 1, 22, 1, 30, 456000),
                         tasks[0].completionDateTime())
        self.failUnless(tasks[0].completed())

    def testNoCompletionDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task />
        </tasks>\n''')
        self.assertEqual(date.DateTime(), tasks[0].completionDateTime())

    def testCompletionDateTimeWithoutTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task completiondate="2005-01-01"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 1, 1, 23, 59, 59, 999999),
                         tasks[0].completionDateTime())
        self.failUnless(tasks[0].completed())

    def testEmptyFontDescription(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task font=""/>
        </tasks>\n''')
        self.assertEqual(None, tasks[0].font())
     
    def testHelveticaMacFont(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task font="0;11;70;90;90;0;Helvetica Neue Light;0"/>
        </tasks>\n''')
        size = tasks[0].font().GetPointSize()
        if operating_system.isMac():  # pragma: no cover
            self.assertEqual(11, size)
        else:  # pragma: no cover
            self.failUnless(size > 0)
     
    def testSans9LinuxFont(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task font="Sans 9"/>
        </tasks>\n''')
        if operating_system.isMac():  # pragma: no cover
            self.assertEqual(None, tasks[0].font())
        else:  # pragma: no cover
            expectedFontSize = 9 if operating_system.isGTK() else 8
            self.assertEqual(expectedFontSize, tasks[0].font().GetPointSize())


class XMLReaderVersion31Test(XMLReaderTestCase):
    tskversion = 31  # New in release 1.2.0.
 
    def writeAndReadTasks(self, *args, **kwargs):
        tasks = super(XMLReaderVersion31Test, self).writeAndReadTasks(*args, 
                                                                      **kwargs)
        tasksById = dict()
     
        def collectIds(tasks):
            for eachTask in tasks:
                tasksById[eachTask.id()] = eachTask
                collectIds(eachTask.children())
             
        collectIds(tasks)
        return tasksById
 
    def assertDepends(self, prerequisite, dependency):
        self.failUnless(prerequisite in dependency.prerequisites())
        self.failUnless(dependency in prerequisite.dependencies())
     
    def testOnePrerequisite(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="1"/>
            <task id="2" prerequisites="1"/>
        </tasks>\n''')
        self.assertDepends(tasks['1'], tasks['2'])
     
    def testTwoPrerequisites(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="1"/>
            <task id="2" prerequisites="1 3"/>
            <task id="3"/>
        </tasks>\n''')
        self.assertDepends(tasks['1'], tasks['2'])
        self.assertDepends(tasks['3'], tasks['2'])
     
    def testChainOfPrerequisites(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="1"/>
            <task id="2" prerequisites="1"/>
            <task id="3" prerequisites="2"/>
        </tasks>\n''')
        self.assertDepends(tasks['1'], tasks['2'])
        self.assertDepends(tasks['2'], tasks['3'])

    def testSubTaskPrerequisite(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="1">
                <task id="1.1"/>
            </task>
            <task id="2" prerequisites="1.1"/>
        </tasks>\n''')
        self.assertDepends(tasks['1.1'], tasks['2'])
     
    def testInterSubTaskPrerequisite(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="1">
                <task id="1.1"/>
                <task id="1.2" prerequisites="1.1"/>
            </task>
        </tasks>\n''')
        self.assertDepends(tasks['1.1'], tasks['1.2'])

    def testMutualPrerequisites(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="1" prerequisites="2"/>
            <task id="2" prerequisites="1"/>
        </tasks>\n''')
        self.assertDepends(tasks['1'], tasks['2'])
        self.assertDepends(tasks['2'], tasks['1'])

    def testMutualPrerequisiteWithChild(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="1" prerequisites="1.1">
                <task id="1.1" prerequisites="1"/>
            </task>
        </tasks>\n''')
        self.assertDepends(tasks['1'], tasks['1.1'])
        self.assertDepends(tasks['1.1'], tasks['1'])
             
    def testSelfPrerequisite(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task id="1" prerequisites="1"/>
        </tasks>\n''')
        self.assertDepends(tasks['1'], tasks['1'])


class XMLReaderVersion33Test(XMLReaderTestCase):
    tskversion = 33  # New in release 1.2.24.
 
    def testReminderBeforeSnooze(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task reminder="2004-01-01 10:00:00" 
                  reminderBeforeSnooze="2004-01-01 9:00:00"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2004, 1, 1, 9, 0, 0), 
                         tasks[0].reminder(includeSnooze=False))
        self.assertEqual(date.DateTime(2004, 1, 1, 10, 0, 0), 
                         tasks[0].reminder())

    def testReminder(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task reminder="2004-01-01 10:00:00"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2004, 1, 1, 10, 0, 0), 
                         tasks[0].reminder(includeSnooze=False))
        self.assertEqual(date.DateTime(2004, 1, 1, 10, 0, 0), 
                         tasks[0].reminder())

    def testRecurrenceNotBasedOnCompletion(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task><recurrence unit="daily"/></task>
        </tasks>''')
        self.failIf(tasks[0].recurrence().recurBasedOnCompletion)

    def testRecurrenceBasedOnCompletion(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <recurrence unit="daily" recurBasedOnCompletion="True"/>
            </task>
        </tasks>''')
        self.failUnless(tasks[0].recurrence().recurBasedOnCompletion)


class XMLReaderVersion34Test(XMLReaderTestCase):
    tskversion = 34  # New in release 1.3.5.
 
    def testPlannedStartDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task plannedstartdate="2005-04-17 10:05:11"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 4, 17, 10, 5, 11),
                         tasks[0].plannedStartDateTime())

    def testPlannedStartDateTimeWithoutTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task plannedstartdate="2005-04-17"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 4, 17), 
                         tasks[0].plannedStartDateTime())

    def testNoPlannedStartDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task />
        </tasks>\n''')
        self.assertEqual(date.DateTime(), tasks[0].plannedStartDateTime())

    def testPlannedStartDateTimeWithMicroseconds(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task plannedstartdate="2005-01-01 22:01:30.456"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 1, 1, 22, 1, 30, 456),
                         tasks[0].plannedStartDateTime())

    def testActualStartDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task actualstartdate="2005-04-17 10:05:11"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 4, 17, 10, 5, 11),
                         tasks[0].actualStartDateTime())

    def testActualStartDateTimeWithoutTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task actualstartdate="2005-04-17"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 4, 17), 
                         tasks[0].actualStartDateTime())

    def testNoActualStartDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task />
        </tasks>\n''')
        self.assertEqual(date.DateTime(), tasks[0].actualStartDateTime())

    def testActualStartDateTimeWithMicroseconds(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task actualstartdate="2005-01-01 22:01:30.456"/>
        </tasks>\n''')
        self.assertEqual(date.DateTime(2005, 1, 1, 22, 1, 30, 456),
                         tasks[0].actualStartDateTime())
     
    def testTaskWithNoteWithCategory(self):
        tasks, categories = self.writeAndReadTasksAndCategories('''
        <tasks>
            <task>
                <note subject="note" id="note1"/>
            </task>
            <category id="cat1" categorizables="note1"/>
        </tasks>\n''')
        self.assertEqual(list(tasks[0].notes())[0], 
                         list(categories[0].categorizables())[0])


class XMLReaderVersion35Test(XMLReaderTestCase):
    tskversion = 35  # New in release 1.3.19.

    def testRecurrenceStopDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task>
                <recurrence stop_datetime="2000-01-10 13:13:13"/>
            </task>
        </tasks>''')
        self.assertEqual(date.DateTime(2000, 1, 10, 13, 13, 13), 
                         tasks[0].recurrence().stop_datetime)
     
    def testCreationDateTimeIsSetToMinDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task/>
        </tasks>''')
        self.assertEqual(date.DateTime.min, tasks[0].creationDateTime())


class XMLReaderVersion36Test(XMLReaderTestCase):
    tskversion = 36  # New in release 1.3.21
 
    def testCreationDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task creationDateTime="2012-12-12 12:00:00.12345"/>
        </tasks>''')
        self.assertEqual(date.DateTime(2012, 12, 12, 12, 0, 0, 12345),
                         tasks[0].creationDateTime())
        
        
class XMLReaderVersion37Test(XMLReaderTestCase):
    tskversion = 37  # New in release 1.3.23
    
    def testModificationDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task modificationDateTime="2012-12-12 12:00:00.12345"/>
        </tasks>''')
        self.assertEqual(date.DateTime(2012, 12, 12, 12, 0, 0, 12345),
                         tasks[0].modificationDateTime())
     
    def testModificationDateTimeWithOtherAttributes(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
            <task modificationDateTime="2012-12-12 12:00:00.12345"/>
        </tasks>''')
        self.assertEqual(date.DateTime(2012, 12, 12, 12, 0, 0, 12345),
                         tasks[0].modificationDateTime())

    def testAdjustDueDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
          <task duedate="2012-12-12 23:59:00:000000" />
        </tasks>''')
        self.assertEqual(date.DateTime(2012, 12, 12, 23, 59, 59, 999999),
                         tasks[0].dueDateTime())

    def test_dontAdjustPlannedStartDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
          <task plannedstartdate="2012-12-12 23:59:00:000000" />
        </tasks>''')
        self.assertEqual(date.DateTime(2012, 12, 12, 23, 59, 0, 0),
                         tasks[0].plannedStartDateTime())

    def test_dontAdjustActualStartDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
          <task actualstartdate="2012-12-12 23:59:00:000000" />
        </tasks>''')
        self.assertEqual(date.DateTime(2012, 12, 12, 23, 59, 0, 0),
                         tasks[0].actualStartDateTime())

    def test_dontAdjustCompletionDateTime(self):
        tasks = self.writeAndReadTasks('''
        <tasks>
          <task completiondate="2012-12-12 23:59:00:000000" />
        </tasks>''')
        self.assertEqual(date.DateTime(2012, 12, 12, 23, 59, 0, 0),
                         tasks[0].completionDateTime())

    def testTaskNoteCategory(self):
        categories = self.writeAndReadCategories('''
        <tasks>
          <task>
            <note id="noteid" />
          </task>
          <category categorizables="noteid" />
        </tasks>''')
        self.assert_('noteid' in [obj.id() for obj in categories[0].categorizables()])

    def testSubtaskNoteCategory(self):
        categories = self.writeAndReadCategories('''
        <tasks>
          <task>
            <task>
              <note id="noteid" />
            </task>
          </task>
          <category categorizables="noteid" />
        </tasks>''')
        self.assert_('noteid' in [obj.id() for obj in categories[0].categorizables()])

    def testCategoryNoteCategory(self):
        categories = self.writeAndReadCategories('''
        <tasks>
          <category categorizables="noteid">
            <note id="noteid" />
          </category>
        </tasks>''')
        self.assert_('noteid' in [obj.id() for obj in categories[0].categorizables()])

    def testSubcategoryNoteCategory(self):
        categories = self.writeAndReadCategories('''
        <tasks>
          <category categorizables="noteid">
            <category>
              <note id="noteid" />
            </category>

          </category>
        </tasks>''')
        self.assert_('noteid' in [obj.id() for obj in categories[0].categorizables()])

    def testTaskAttachmentNoteCategory(self):
        categories = self.writeAndReadCategories('''
        <tasks>
          <task>
            <attachment location="test" type="file">
              <note id="noteid" />
            </attachment>
          </task>
          <category categorizables="noteid" />
        </tasks>''')
        self.assert_('noteid' in [obj.id() for obj in categories[0].categorizables()])

    def testSubtaskAttachmentNoteCategory(self):
        categories = self.writeAndReadCategories('''
        <tasks>
          <task subject="Task">
            <task subject="Subtask">
              <attachment location="test" type="file" subject="Attachment">
                <note id="noteid" subject="Note" />
              </attachment>
            </task>
          </task>
          <category categorizables="noteid" subject="Category" />
        </tasks>''')
        self.assert_('noteid' in [obj.id() for obj in categories[0].categorizables()])



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>
Copyright (C) 2008 Thomas Sonne Olesen <tpo@sonnet.dk>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
from taskcoachlib.domain import date, base, task
from taskcoachlib.thirdparty.pubsub import pub
import base as baseeffort
import weakref


class Effort(baseeffort.BaseEffort, base.Object):
    def __init__(self, task=None, start=None, stop=None, *args, **kwargs):
        super(Effort, self).__init__(task, start or date.DateTime.now(), stop, 
            *args, **kwargs)
        self.__updateDurationCache()
        
    def setTask(self, task):
        if self._task is None: 
            # We haven't been fully initialised yet, so allow setting of the
            # task, without notifying observers. Also, don't call addEffort()
            # on the new task, because we assume setTask was invoked by the
            # new task itself.
            self._task = None if task is None else weakref.ref(task)
            return
        if task in (self.task(), None): 
            # command.PasteCommand may try to set the parent to None
            return
        event = patterns.Event()  # Change monitor needs one event to detect task change
        self._task().removeEffort(self)
        self._task = weakref.ref(task)
        self._task().addEffort(self)
        event.send()
        pub.sendMessage(self.taskChangedEventType(), newValue=task, sender=self)
        
    setParent = setTask  # FIXME: should we create a common superclass for Effort and Task?

    @classmethod
    def monitoredAttributes(class_):
        return base.Object.monitoredAttributes() + ['start', 'stop']

    def task(self):
        return None if self._task is None else self._task()

    @classmethod
    def taskChangedEventType(class_):
        return 'pubsub.effort.task'
    
    def __str__(self):
        return 'Effort(%s, %s, %s)' % (self.task(), self._start, self._stop)
    
    __repr__ = __str__
        
    def __getstate__(self):
        state = super(Effort, self).__getstate__()
        state.update(dict(task=self.task(), start=self._start, stop=self._stop))
        return state

    @patterns.eventSource
    def __setstate__(self, state, event=None):
        super(Effort, self).__setstate__(state, event=event)
        self.setTask(state['task'])
        self.setStart(state['start'])
        self.setStop(state['stop'])

    def __getcopystate__(self):
        state = super(Effort, self).__getcopystate__()
        state.update(dict(task=self.task(), start=self._start, stop=self._stop))
        return state
   
    def duration(self, now=date.DateTime.now):
        return now() - self._start if self.__cachedDuration is None else self.__cachedDuration
     
    def setStart(self, startDateTime):
        if startDateTime == self._start:
            return
        self._start = startDateTime
        self.__updateDurationCache()
        pub.sendMessage(self.startChangedEventType(), newValue=startDateTime,
                        sender=self)
        self.task().sendTimeSpentChangedMessage()
        self.sendDurationChangedMessage()
        if self.task().hourlyFee():
            self.sendRevenueChangedMessage()

    @classmethod
    def startChangedEventType(class_):
        return 'pubsub.effort.start'

    def setStop(self, newStop=None):
        if newStop is None:
            newStop = date.DateTime.now()
        elif newStop == date.DateTime.max:
            newStop = None
        if newStop == self._stop:
            return
        previousStop = self._stop
        self._stop = newStop
        self.__updateDurationCache()
        if newStop == None:
            pub.sendMessage(self.trackingChangedEventType(), newValue=True, 
                            sender=self)
            self.task().sendTrackingChangedMessage(tracking=True)
        elif previousStop == None:
            pub.sendMessage(self.trackingChangedEventType(), newValue=False,
                            sender=self)
            self.task().sendTrackingChangedMessage(tracking=False)
        self.task().sendTimeSpentChangedMessage()
        pub.sendMessage(self.stopChangedEventType(), newValue=self._stop,
                           sender=self)
        self.sendDurationChangedMessage()
        if self.task().hourlyFee():
            self.sendRevenueChangedMessage()
            
    @classmethod
    def stopChangedEventType(class_):
        return 'pubsub.effort.stop'
        
    def __updateDurationCache(self):
        self.__cachedDuration = self._stop - self._start if self._stop else None
        
    def isBeingTracked(self, recursive=False):  # pylint: disable=W0613
        return self._stop is None

    def revenue(self):
        return self.duration().hours() * self.task().hourlyFee()

    @staticmethod
    def periodSortFunction(**kwargs):
        # Sort by start of effort first, then make sure the Total entry comes
        # first and finally sort by task subject:
        return lambda effort: (effort.getStart(), effort.isTotal(),
                               effort.task().subject(recursive=True))
    
    @classmethod
    def periodSortEventTypes(class_):
        ''' The event types that influence the effort sort order. '''
        return (class_.startChangedEventType(), class_.taskChangedEventType(),
                task.Task.subjectChangedEventType())
            
    @classmethod    
    def modificationEventTypes(class_):
        eventTypes = super(Effort, class_).modificationEventTypes()
        return eventTypes + [class_.taskChangedEventType(), 
                             class_.startChangedEventType(), 
                             class_.stopChangedEventType()]



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns
import dateandtime, timedelta
import logging
import timedelta
import weakref
import bisect


class ScheduledMethod(object):
    def __init__(self, method):
        self.__func = method.im_func
        self.__self = weakref.ref(method.im_self)

    def __eq__(self, other):
        return self.__func is other.__func and self.__self() is other.__self()

    def __hash__(self):
        return hash(self.__dict__['_ScheduledMethod__func'])

    def __call__(self, *args, **kwargs):
        obj = self.__self()
        if obj is None:
            Scheduler().unschedule(self)
        else:
            self.__func(obj, *args, **kwargs)


class TwistedScheduler(object):
    """
    A class to schedule jobs at specified date/time. Unlike apscheduler, this
    uses Twisted instead of threading, in order to avoid busy waits.
    """
    def __init__(self):
        super(TwistedScheduler, self).__init__()
        self.__jobs = []
        self.__nextCall = None
        self.__firing = False

    def __schedule(self, job, dateTime, interval):
        if self.__nextCall is not None:
            self.__nextCall.cancel()
            self.__nextCall = None
        bisect.insort_right(self.__jobs, (dateTime, job, interval))
        if not self.__firing:
            self.__fire()

    def scheduleDate(self, job, dateTime):
        """
        Schedules 'job' to be called at 'dateTime'. This assumes the caller is the
        main loop thread.
        """
        self.__schedule(job, dateTime, None)

    def scheduleInterval(self, job, interval, startDateTime=None):
        self.__schedule(job, startDateTime or dateandtime.Now() + interval, interval)

    def unschedule(self, theJob):
        for idx, (ts, job, interval) in enumerate(self.__jobs):
            if job == theJob:
                del self.__jobs[idx]
                break

    def isScheduled(self, theJob):
        for ts, job, interval in self.__jobs:
            if job == theJob:
                return True
        return False

    def shutdown(self):
        if self.__nextCall is not None:
            self.__nextCall.cancel()
            self.__nextCall = None
        self.__jobs = []

    def jobs(self):
        return [job for ts, job, interval in self.__jobs]

    def __fire(self):
        self.__firing = True
        try:
            while self.__jobs and self.__jobs[0][0] <= dateandtime.Now():
                ts, job, interval = self.__jobs.pop(0)
                try:
                    job()
                except:
                    # Hum.
                    import traceback
                    traceback.print_exc()
                if interval is not None:
                    self.__schedule(job, ts + interval, interval)
        finally:
            self.__firing = False

        if self.__jobs and self.__nextCall is None:
            dt = self.__jobs[0][0] - dateandtime.Now()
            nextDuration = int((dt.microseconds + (dt.seconds + dt.days * 24 * 3600) * 10**6) / 10**3)
            nextDuration = max(nextDuration, 1)
            nextDuration = min(nextDuration, 2**31-1)
            from twisted.internet import reactor
            self.__nextCall = reactor.callLater(1.0 * nextDuration / 1000, self.__callback)

    def __callback(self):
        self.__nextCall = None
        self.__fire()


class Scheduler(object):
    __metaclass__ = patterns.Singleton

    def __init__(self, *args, **kwargs):
        super(Scheduler, self).__init__(*args, **kwargs)
        self.__scheduler = TwistedScheduler()

    def shutdown(self):
        self.__scheduler.shutdown()

    def schedule(self, function, dateTime):
        job = ScheduledMethod(function)
        self.__scheduler.scheduleDate(job, dateTime)
        return job

    def schedule_interval(self, function, days=0, minutes=0, seconds=0):
        job = ScheduledMethod(function)
        if not self.__scheduler.isScheduled(job):
            startDate = dateandtime.Now().endOfDay() if days > 0 else None
            self.__scheduler.scheduleInterval(job, timedelta.TimeDelta(days=days, minutes=minutes, seconds=seconds), startDateTime=startDate)
            return job

    def unschedule(self, function):
        job = function if isinstance(function, ScheduledMethod) else ScheduledMethod(function)
        self.__scheduler.unschedule(job)

    def is_scheduled(self, function):
        return self.__scheduler.isScheduled(ScheduledMethod(function))

    def get_jobs(self):
        return self.__scheduler.jobs()



# -*- coding: UTF-8 -*-
#This is generated code - do not edit
encoding = 'UTF-8'
dict = {'Cut': 'Ebaki', 'New task': 'Zeregin berria', 'No': 'Ez', 'Monthly': 'Hilero', 'Priority': 'Lehentasuna', 'Editor': 'Editorea', 'Files': 'Fitxategiak', 'Description': 'Deskribapena', 'Never': 'Inoiz ez', 'New attachment': 'Eranskin berria', 'Copy': 'Kopiatu', 'Paste': 'Itsatsi', 'Delete': 'Ezabatu', 'Language': 'Hizkuntza', 'Always': 'Beti', 'Notes': 'Oharrak', 'Daily': 'Egunero', "Couldn't load settings from TaskCoach.ini:\n%s": 'Ezin da TaskCoach.ini-tik konfigurazioa kargatu:\n%s', 'Attachments': 'Eranskinak', 'Yearly': 'Urtero', '%s file error': '%s fitxategi akatsa', 'Due date': 'Iraungipen-data', 'Insert a new task': 'Txertatu zeregin berria', 'Yes': 'Bai', 'Weekly': 'Astero', 'Subject': 'Gaia'}


'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import patterns, command
from taskcoachlib.domain import task, date
from CommandTestCase import CommandTestCase


    
class DeleteCommandTest(CommandTestCase):
    def setUp(self):
        super(DeleteCommandTest, self).setUp()
        self.item = task.Task()
        self.items = patterns.List([self.item])
        
    def deleteItem(self, items=None):
        delete = command.DeleteCommand(self.items, items or [])
        delete.do()
        
    def testDeleteItem_WithoutSelection(self):
        self.deleteItem()
        self.assertDoUndoRedo(lambda: self.assertEqual([self.item], self.items))
        
    def testDeleteItem_WithSelection(self):
        self.deleteItem([self.item])
        self.assertDoUndoRedo(lambda: self.assertEqual([], self.items),
                              lambda: self.assertEqual([self.item], self.items))

    def testItemsAreNotNew(self):
        self.failIf(command.DeleteCommand(self.items, []).items_are_new())


class EditSubjectTestCase(CommandTestCase):
    ItemClass = task.Task
    ContainerClass = task.TaskList
    
    def setUp(self):
        super(EditSubjectTestCase, self).setUp()
        self.item1 = self.ItemClass(subject='item1')
        self.item2 = self.ItemClass(subject='item2')
        self.container = self.ContainerClass([self.item1, self.item2])
        
    def editSubject(self, newSubject, *items):
        editSubjectCommand = command.EditSubjectCommand(self.container, 
                                                        items, 
                                                        newValue=newSubject)
        editSubjectCommand.do()
        
    def testEditSubject(self):
        self.editSubject('new', self.item1)
        self.assertDoUndoRedo(lambda: self.assertEqual('new', self.item1.subject()),
                              lambda: self.assertEqual('item1', self.item1.subject()))
        
    def testEditMultipleSubjects(self):
        self.editSubject('new', self.item1, self.item2)
        self.assertDoUndoRedo(lambda: self.assertEqual('newnew', 
                                      self.item1.subject() + self.item2.subject()),
                              lambda: self.assertEqual('item1item2', 
                                      self.item1.subject() + self.item2.subject()))

    def testItemsAreNotNew(self):
        self.failIf(command.EditSubjectCommand(self.container, [], 
                    newValue='New subject').items_are_new())

    def testModificationDateTime(self):
        self.editSubject('new', self.item1)
        self.assertDoUndoRedo(lambda: self.failUnless(self.item1.modificationDateTime() > date.DateTime.min),
                              lambda: self.assertEqual(date.DateTime.min, self.item1.modificationDateTime()))


class EditDescriptionTestCase(CommandTestCase):
    ItemClass = task.Task
    ContainerClass = task.TaskList
    
    def setUp(self):
        super(EditDescriptionTestCase, self).setUp()
        self.item1 = self.ItemClass(description='item1')
        self.item2 = self.ItemClass(description='item2')
        self.container = self.ContainerClass([self.item1, self.item2])
        
    def edit_description(self, new_description, *items):
        edit_subject = command.EditDescriptionCommand(self.container, items, 
                                                      newValue=new_description)
        edit_subject.do()
        
    def testEditSubject(self):
        self.edit_description('new', self.item1)
        self.assertDoUndoRedo(lambda: self.assertEqual('new', self.item1.description()),
                              lambda: self.assertEqual('item1', self.item1.description()))
        
    def testEditMultipleDescriptions(self):
        self.edit_description('new', self.item1, self.item2)
        self.assertDoUndoRedo(lambda: self.assertEqual('newnew', 
                                      self.item1.description() + self.item2.description()),
                              lambda: self.assertEqual('item1item2', 
                                      self.item1.description() + self.item2.description()))

    def testItemsAreNotNew(self):
        self.failIf(command.EditDescriptionCommand(self.container, [], 
                    newValue='New description').items_are_new())

    def testModificationDateTime(self):
        self.edit_description('new', self.item1)
        self.assertDoUndoRedo(lambda: self.failUnless(self.item1.modificationDateTime() > date.DateTime.min),
                              lambda: self.assertEqual(date.DateTime.min, self.item1.modificationDateTime()))



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from win32file import *
from win32con import *
from win32event import *
import os, threading
from taskcoachlib.filesystem import base


class DirectoryWatcher(object):
    ADDED       = 1
    REMOVED     = 2
    MODIFIED    = 3
    RENAMED_OLD = 4
    RENAMED_NEW = 5

    def __init__(self, path):
        super(DirectoryWatcher, self).__init__()

        self.dirHandle = CreateFile(path,
                                    GENERIC_READ,
                                    FILE_SHARE_READ|FILE_SHARE_DELETE|FILE_SHARE_WRITE,
                                    None,
                                    OPEN_EXISTING,
                                    FILE_FLAG_BACKUP_SEMANTICS|FILE_FLAG_OVERLAPPED,
                                    0)
        self.buffer = AllocateReadBuffer(8192)
        self.overlapped = OVERLAPPED()
        self.overlapped.hEvent = CreateEvent(None, False, False, None)

    def wait(self, recurse=False, timeout=INFINITE):
        ReadDirectoryChangesW(self.dirHandle, self.buffer, recurse,
                              FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_DIR_NAME | \
                              FILE_NOTIFY_CHANGE_ATTRIBUTES | FILE_NOTIFY_CHANGE_SIZE | \
                              FILE_NOTIFY_CHANGE_LAST_WRITE,
                              self.overlapped)

        rc = WaitForSingleObject(self.overlapped.hEvent, timeout)
        if rc == WAIT_OBJECT_0:
            try:
                size = GetOverlappedResult(self.dirHandle, self.overlapped, True)
            except Exception, e:
                if e.args[0] == 995:
                    return None
                raise
            if size == 0:
                return None
            else:
                return FILE_NOTIFY_INFORMATION(self.buffer, size)

    def stop(self):
        CloseHandle(self.dirHandle)


class FilesystemNotifier(base.NotifierBase):
    def __init__(self):
        super(FilesystemNotifier, self).__init__()

        self.watcher = None
        self.thread = None
        self.lock = threading.Lock()

    def setFilename(self, filename):
        self.lock.acquire()
        try:
            if self.watcher is not None:
                self.watcher.stop()
                self.thread.join()
                self.watcher = None
                self.thread = None
            super(FilesystemNotifier, self).setFilename(filename)
            if self._filename:
                self.watcher = DirectoryWatcher(self._path)
                self.thread = threading.Thread(target=self._run)
                self.thread.setDaemon(True)
                self.thread.start()
        finally:
            self.lock.release()

    def stop(self):
        self.lock.acquire()
        try:
            if self.watcher is not None:
                self.watcher.stop()
                self.thread.join()
                self.watcher = None
                self.thread = None
        finally:
            self.lock.release()

    def saved(self):
        with self.lock:
            super(FilesystemNotifier, self).saved()

    def onFileChanged(self):
        raise NotImplementedError

    def _run(self):
        while True:
            self.lock.acquire()
            try:
                watcher, myname = self.watcher, self._filename
            finally:
                self.lock.release()

            if watcher is not None:
                changes = watcher.wait()
                if changes is None:
                    return
                for change, name in changes:
                    if name == os.path.split(myname)[-1]:
                        if self._check(myname) and myname:
                            self.stamp = os.stat(myname).st_mtime
                            self.onFileChanged()
                            break



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import gui, config, persistence
from taskcoachlib.domain import note, attachment, category


class NoteViewerTest(test.wxTestCase):
    def setUp(self):
        super(NoteViewerTest, self).setUp()
        self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.note = note.Note()
        self.taskFile.notes().append(self.note)
        self.viewer = gui.viewer.NoteViewer(self.frame, self.taskFile, 
                                            self.settings, 
                                            notesToShow=self.taskFile.notes())

    def tearDown(self):
        super(NoteViewerTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def firstItem(self):
        widget = self.viewer.widget
        return widget.GetFirstChild(widget.GetRootItem())[0]

    def firstItemText(self, column=0):
        return self.viewer.widget.GetItemText(self.firstItem(), column)

    def firstItemIcon(self, column=0):    
        return self.viewer.widget.GetItemImage(self.firstItem(), column=column)

    def testLocalNoteViewerForItemWithoutNotes(self):
        localViewer = gui.viewer.NoteViewer(self.frame, self.taskFile, 
                                            self.settings, 
                                            notesToShow=note.NoteContainer())
        self.failIf(localViewer.presentation())
        
    def testShowDescriptionColumn(self):
        self.note.setDescription('Description')
        self.viewer.showColumnByName('description')
        self.assertEqual('Description', self.firstItemText(column=1))

    def testShowCategoriesColumn(self):
        newCategory = category.Category('Category')
        self.taskFile.categories().append(newCategory)
        self.note.addCategory(newCategory)
        newCategory.addCategorizable(self.note)
        self.viewer.showColumnByName('categories')
        self.assertEqual('Category', self.firstItemText(column=3))
        
    def testShowAttachmentColumn(self):
        self.note.addAttachments(attachment.FileAttachment('whatever'))
        self.assertEqual(self.viewer.imageIndex['paperclip_icon'], 
                         self.firstItemIcon(column=2))

    def testFilterOnAllCategories(self):
        cat1 = category.Category('category 1')
        cat2 = category.Category('category 2')
        self.note.addCategory(cat1)
        cat1.addCategorizable(self.note)
        self.taskFile.categories().extend([cat1, cat2])
        cat1.setFiltered(True)
        cat2.setFiltered(True)
        self.assertEqual(1, self.viewer.size())
        self.settings.setboolean('view', 'categoryfiltermatchall', True)
        self.assertEqual(0, self.viewer.size())
        
    def testFilterOnAnyCategory(self):
        cat1 = category.Category('category 1')
        cat2 = category.Category('category 2')
        self.note.addCategory(cat1)
        cat1.addCategorizable(self.note)
        self.taskFile.categories().extend([cat1, cat2])
        cat1.setFiltered(True)
        cat2.setFiltered(True)
        self.assertEqual(1, self.viewer.size())
        self.settings.setboolean('view', 'categoryfiltermatchall', False)
        self.assertEqual(1, self.viewer.size())



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
import test
from taskcoachlib import gui, config, persistence, operating_system
from taskcoachlib.domain import category, attachment


class DummyEvent(object):
    def Skip(self): # pragma: no cover
        pass
    

class CategoryEditorTest(test.wxTestCase):
    def setUp(self):
        super(CategoryEditorTest, self).setUp()
        self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.categories = self.taskFile.categories()
        self.categories.extend(self.createCategories())
        self.editor = gui.dialog.editor.CategoryEditor(self.frame, 
            list(self.categories), self.settings, self.categories, 
            self.taskFile)

    def tearDown(self):
        # CategoryEditor uses CallAfter for setting the focus, make sure those 
        # calls are dealt with, otherwise they'll turn up in other tests
        if operating_system.isGTK():
            wx.Yield() # pragma: no cover 
        super(CategoryEditorTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    # pylint: disable=E1101,E1103,W0212
    
    def createCategories(self):
        # pylint: disable=W0201
        self.category = category.Category('Category to edit')
        self.attachment = attachment.FileAttachment('some attachment')
        self.category.addAttachments(self.attachment)
        return [self.category]

    def setSubject(self, newSubject):
        page = self.editor._interior[0]
        page._subjectEntry.SetFocus()
        page._subjectEntry.SetValue(newSubject)
        if operating_system.isGTK(): # pragma: no cover 
            page._subjectSync.onAttributeEdited(DummyEvent())
        else: # pragma: no cover 
            page._descriptionEntry.SetFocus()

    def setDescription(self, newDescription):
        page = self.editor._interior[0]
        page._descriptionEntry.SetFocus()
        page._descriptionEntry.SetValue(newDescription)
        if operating_system.isGTK(): # pragma: no cover 
            page._descriptionSync.onAttributeEdited(DummyEvent())
        else:  # pragma: no cover
            page._subjectEntry.SetFocus()
        
    def testCreate(self):
        self.assertEqual('Category to edit', self.editor._interior[0]._subjectEntry.GetValue())
    
    def testEditSubject(self):
        self.setSubject('Done')
        self.assertEqual('Done', self.category.subject())

    def testEditDescription(self):
        self.setDescription('Description')
        self.assertEqual('Description', self.category.description())        

    def testAddAttachment(self):
        self.editor._interior[2].viewer.onDropFiles(self.category, ['filename'])
        self.failUnless('filename' in [att.location() for att in self.category.attachments()])
        self.failUnless('filename' in [att.subject() for att in self.category.attachments()])
        
    def testRemoveAttachment(self):
        self.editor._interior[2].viewer.select(self.category.attachments())
        self.editor._interior[2].viewer.deleteItemCommand().do()
        self.assertEqual([], self.category.attachments())

    def testEditMutualExclusiveSubcategories(self):
        self.editor._interior[0]._exclusiveSubcategoriesCheckBox.SetValue(True)
        self.editor._interior[0]._exclusiveSubcategoriesSync.onAttributeEdited(DummyEvent())
        self.failUnless(self.category.hasExclusiveSubcategories())
        
    def testAddNote(self):
        viewer = self.editor._interior[1].viewer
        viewer.newItemCommand(viewer.presentation()).do() 
        self.assertEqual(1, len(self.category.notes()))



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from object import Object, CompositeObject, SynchronizedObject
from attribute import Attribute, SetAttribute
from collection import Collection
from filter import Filter, SearchFilter, SelectedItemsFilter, DeletedFilter
from sorter import Sorter, TreeSorter
from owner import DomainObjectOwnerMetaclass



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import datetime, timedelta, re, time
from .date import Date
from .fix import StrftimeFix


class DateTime(StrftimeFix, datetime.datetime):
    secondsPerMinute = 60
    minutesPerHour = 60
    hoursPerDay = 24
    secondsPerHour = minutesPerHour * secondsPerMinute
    secondsPerDay = hoursPerDay * secondsPerHour

    def __new__(class_, *args, **kwargs):
        if not args and not kwargs:
            max = datetime.datetime.max # pylint: disable=W0622
            args = (max.year, max.month, max.day, 
                    max.hour, max.minute, max.second, max.microsecond)
        return datetime.datetime.__new__(class_, *args, **kwargs)

    @staticmethod
    def fromDateTime(dateTime):
        return DateTime(year=dateTime.year, month=dateTime.month, day=dateTime.day,
                        hour=dateTime.hour, minute=dateTime.minute, second=dateTime.second,
                        microsecond=dateTime.microsecond)

    def date(self):
        return Date(self.year, self.month, self.day)

    def weeknumber(self):
        return self.isocalendar()[1]

    def weekday(self):
        return self.isoweekday() # Sunday = 7, Monday = 1, etc.
    
    def toordinal(self):
        ''' Return the ordinal number of the day, plus a fraction between 0 and
            1 for parts of the day. '''
        ordinal = super(DateTime, self).toordinal()
        seconds = self.hour * self.secondsPerHour + \
                  self.minute * self.secondsPerMinute + \
                  self.second
        return ordinal + (seconds / float(self.secondsPerDay))
        
    def startOfDay(self):
        return self.replace(hour=0, minute=0, second=0, microsecond=0)
        
    def endOfDay(self):
        return self.replace(hour=23, minute=59, second=59, microsecond=999999)

    def startOfWeek(self):
        days = self.weekday()
        monday = self - timedelta.TimeDelta(days=days-1)
        return DateTime(monday.year, monday.month, monday.day)
        
    def endOfWeek(self):
        days = self.weekday()
        sunday = self + timedelta.TimeDelta(days=7-days)
        return DateTime(sunday.year, sunday.month, sunday.day).endOfDay()
    
    def startOfWorkWeek(self):
        days = self.weekday()
        monday = self - timedelta.TimeDelta(days=days-1)
        return DateTime(monday.year, monday.month, monday.day)
    
    def endOfWorkWeek(self):
        days = 5 - self.weekday()
        if days < 0:
            days += 7
        friday = self + timedelta.TimeDelta(days=days)
        return DateTime(friday.year, friday.month, friday.day).endOfDay()
            
    def startOfMonth(self):
        return DateTime(self.year, self.month, 1)
        
    def endOfMonth(self):
        for lastday in [31,30,29,28]:
            try:
                return DateTime(self.year, self.month, lastday).endOfDay()
            except ValueError:
                pass

    def startOfYear(self):
        return DateTime(self.year, 1, 1).startOfDay()
    
    def endOfYear(self):
        return DateTime(self.year, 12, 31).endOfDay()
                
    def __sub__(self, other):
        ''' Make sure substraction returns instances of the right classes. '''
        if self == DateTime() and isinstance(other, datetime.datetime):
            max = timedelta.TimeDelta.max # pylint: disable=W0622
            return timedelta.TimeDelta(max.days, max.seconds, max.microseconds)
        result = super(DateTime, self).__sub__(other)
        if isinstance(result, datetime.timedelta):
            result = timedelta.TimeDelta(result.days, result.seconds, 
                                         result.microseconds)
        elif isinstance(result, datetime.datetime):
            result = self.__class__(result.year, result.month, result.day, 
                                    result.hour, result.minute, result.second, 
                                    result.microsecond)
        return result

    def __add__(self, other):
        result = super(DateTime, self).__add__(other)
        return self.__class__(result.year, result.month, result.day, 
            result.hour, result.minute, result.second, result.microsecond)
        

DateTime.max = DateTime(datetime.datetime.max.year, 12, 31).endOfDay()
DateTime.min = DateTime(datetime.datetime.min.year, 1, 1).startOfDay()


def parseDateTime(string, *timeDefaults):
    if string in ('', 'None'):
        return None
    else:
        args = [int(arg) for arg in re.split('[-:. ]', string)]
        if len(args) == 3:  # We parsed a date, no time
            args.extend(timeDefaults)
        return DateTime(*args)  # pylint: disable=W0142
        

def Now():
    return DateTime.now()

def Today():
    # For backwards compatibility: "Today()" may be used in templates
    return Now().replace(hour=0, minute=0, second=0, microsecond=0)

def Tomorrow():
    return Now() + timedelta.ONE_DAY

def Yesterday():
    return Now() - timedelta.ONE_DAY

def LastDayOfCurrentMonth(localtime=time.localtime):
    now = localtime()
    year, nextMonth = now[0], now[1]+1
    if nextMonth > 12:
        nextMonth = 1
        year += 1
    return DateTime(year, nextMonth, 1) - timedelta.ONE_DAY 




'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import config
from taskcoachlib.gui.dialog import version 
import test


class DummyEvent(object):
    def Skip(self):
        pass


class CommonTestsMixin(object):
    def testCreateAndClose(self):
        self.dialog.onClose(DummyEvent())
        self.failUnless(self.settings.getboolean('version', 'notify'))

    def testNoMoreNotifications(self):
        self.dialog.check.SetValue(False)
        self.dialog.onClose(DummyEvent())
        self.failIf(self.settings.getboolean('version', 'notify'))


class VersionDialogTestCase(test.TestCase):
    def setUp(self):
        self.settings = config.Settings(load=False)
        self.dialog = self.createDialog()
        
    def createDialog(self):
        raise NotImplementedError  # pragma: no cover
        

class NewVersionDialogTest(CommonTestsMixin, VersionDialogTestCase):
    def createDialog(self):
        return version.NewVersionDialog(None, version='0.0', message='', 
                                        settings=self.settings)


class VersionUpToDateDialogTest(CommonTestsMixin, VersionDialogTestCase):
    def createDialog(self):
        return version.VersionUpToDateDialog(None, version='0.0', message='', 
                                             settings=self.settings)
        

class NoVersionDialogTest(CommonTestsMixin, VersionDialogTestCase):
    def createDialog(self):
        return version.NoVersionDialog(None, version='0.0', message='', 
                                       settings=self.settings)
        
        
class PrereleaseVersionDialogTest(CommonTestsMixin, VersionDialogTestCase):
    def createDialog(self):
        return version.PrereleaseVersionDialog(None, version='0.0', message='', 
                                               settings=self.settings)



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import gui, config, persistence, render
from taskcoachlib.domain import task, effort, date
from unittests import dummy
import test
import wx


class EffortViewerUnderTest(gui.viewer.EffortViewer):  # pylint: disable=W0223
    def createWidget(self):
        return dummy.DummyWidget(self)
    
    def columns(self):
        return []


class EffortViewerForSpecificTasksTest(test.wxTestCase):
    def setUp(self):
        super(EffortViewerForSpecificTasksTest, self).setUp()
        self.settings = config.Settings(load=False)
        task.Task.settings = self.settings
        self.taskFile = persistence.TaskFile()
        self.task1 = task.Task('Task 1')
        self.task2 = task.Task('Task 2')
        self.taskFile.tasks().extend([self.task1, self.task2])
        self.effort1 = effort.Effort(self.task1, date.DateTime(2006, 1, 1),
            date.DateTime(2006, 1, 2))
        self.task1.addEffort(self.effort1)
        self.effort2 = effort.Effort(self.task2, date.DateTime(2006, 1, 2),
            date.DateTime(2006, 1, 3))
        self.task2.addEffort(self.effort2)
        self.viewer = EffortViewerUnderTest(self.frame, self.taskFile,  
            self.settings, tasksToShowEffortFor=task.TaskList([self.task1]))

    def tearDown(self):
        super(EffortViewerForSpecificTasksTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def testViewerShowsOnlyEffortForSpecifiedTask(self):
        self.assertEqual([self.effort1], self.viewer.presentation())
        
    def testEffortEditorDoesUseAllTasks(self):
        dialog = self.viewer.newItemDialog()
        self.assertEqual(2, len(dialog._taskFile.tasks()))  # pylint: disable=W0212
        
    def testViewerKeepsShowingOnlyEffortForSpecifiedTasksWhenSwitchingAggregation(self):
        self.settings.settext(self.viewer.settingsSection(), 'aggregation', 
                              'week')
        self.assertEqual(2, len(self.viewer.presentation()))
        
        
class EffortViewerStatusMessageTest(test.wxTestCase):
    def setUp(self):
        super(EffortViewerStatusMessageTest, self).setUp()
        self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.task = task.Task()
        self.taskFile.tasks().append(self.task)
        self.effort1 = effort.Effort(self.task, date.DateTime(2006, 1, 1),
            date.DateTime(2006, 1, 2))
        self.effort2 = effort.Effort(self.task, date.DateTime(2006, 1, 2),
            date.DateTime(2006, 1, 3))
        self.viewer = EffortViewerUnderTest(self.frame, self.taskFile,  
            self.settings)

    def tearDown(self):
        super(EffortViewerStatusMessageTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def assertStatusMessages(self, message1, message2):
        self.assertEqual((message1, message2), self.viewer.statusMessages())
        
    def testStatusMessage_EmptyTaskList(self):
        self.taskFile.tasks().clear()
        self.assertStatusMessages('Effort: 0 selected, 0 visible, 0 total. Time spent: 0:00:00 selected, 0:00:00 visible, 0:00:00 total',
            'Status: 0 tracking')
            
    def testStatusMessage_OneTaskNoEffort(self):
        self.assertStatusMessages('Effort: 0 selected, 0 visible, 0 total. Time spent: 0:00:00 selected, 0:00:00 visible, 0:00:00 total',
            'Status: 0 tracking')
        
    def testStatusMessage_OneTaskOneEffort(self):
        self.task.addEffort(self.effort1)
        self.assertStatusMessages('Effort: 0 selected, 1 visible, 1 total. Time spent: 0:00:00 selected, 24:00:00 visible, 24:00:00 total',
            'Status: 0 tracking')
            
    def testStatusMessage_OneTaskTwoEfforts(self):
        self.task.addEffort(self.effort1)
        self.task.addEffort(self.effort2)
        self.assertStatusMessages('Effort: 0 selected, 2 visible, 2 total. Time spent: 0:00:00 selected, 48:00:00 visible, 48:00:00 total',
            'Status: 0 tracking')
            
    def testStatusMessage_OneTaskOneActiveEffort(self):
        self.task.addEffort(effort.Effort(self.task))
        self.assertStatusMessages('Effort: 0 selected, 1 visible, 1 total. Time spent: 0:00:00 selected, 0:00:00 visible, 0:00:00 total',
            'Status: 1 tracking')
        
    def testStatusMessageInAggregatedMode_OneTaskNoEffort(self):
        self.settings.settext(self.viewer.settingsSection(), 'aggregation', 
                              'day')
        self.assertStatusMessages('Effort: 0 selected, 0 visible, 0 total. Time spent: 0:00:00 selected, 0:00:00 visible, 0:00:00 total',
            'Status: 0 tracking')

    def testStatusMessageInAggregateMode_OneTaskOneEffort(self):
        self.settings.settext(self.viewer.settingsSection(), 'aggregation', 
                              'day')
        self.task.addEffort(self.effort1)
        self.assertStatusMessages('Effort: 0 selected, 2 visible, 1 total. Time spent: 0:00:00 selected, 48:00:00 visible, 24:00:00 total',
            'Status: 0 tracking')

    def testStatusMessageInAggregateMode_OneTaskTwoEfforts(self):
        self.settings.settext(self.viewer.settingsSection(), 'aggregation', 
                              'day')
        self.task.addEffort(self.effort1)
        self.task.addEffort(self.effort2)
        self.assertStatusMessages('Effort: 0 selected, 4 visible, 2 total. Time spent: 0:00:00 selected, 96:00:00 visible, 48:00:00 total',
            'Status: 0 tracking')


class EffortViewerTest(test.wxTestCase):
    def setUp(self):
        super(EffortViewerTest, self).setUp()
        self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.task = task.Task('task')
        self.taskFile.tasks().append(self.task)
        self.effort1 = effort.Effort(self.task, date.DateTime(2006, 1, 1),
            date.DateTime(2006, 1, 2))
        self.effort2 = effort.Effort(self.task, date.DateTime(2006, 1, 2),
            date.DateTime(2006, 1, 3))
        self.viewer = gui.viewer.EffortViewer(self.frame, self.taskFile, 
                                              self.settings)

    def tearDown(self):
        super(EffortViewerTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    @test.skipOnPlatform('__WXMSW__')  # GetItemBackgroundColour doesn't work on Windows
    def testEffortBackgroundColor(self):  # pragma: no cover
        self.task.setBackgroundColor(wx.RED)
        self.task.addEffort(self.effort1)
        self.assertEqual(wx.RED, self.viewer.widget.GetItemBackgroundColour(0))

    @test.skipOnPlatform('__WXMSW__')  # GetItemBackgroundColour doesn't work on Windows
    def testUpdateEffortBackgroundColor(self):  # pragma: no cover
        self.task.addEffort(self.effort1)
        self.task.setBackgroundColor(wx.RED)
        self.assertEqual(wx.RED, self.viewer.widget.GetItemBackgroundColour(0))
    
    def testSearch(self):
        self.task.addEffort(self.effort1)
        self.viewer.presentation().setSearchFilter('no such task')
        self.assertEqual(0, len(self.viewer.presentation()))
        self.viewer.presentation().setSearchFilter(self.task.subject())
        self.assertEqual(1, len(self.viewer.presentation()))
        
    def testSearchIncludeSubitems(self):
        self.task.addEffort(self.effort1)
        child = task.Task('child')
        self.task.addChild(child)
        child.setParent(self.task)
        self.taskFile.tasks().append(child)
        child.addEffort(effort.Effort(child))
        self.assertEqual(2, len(self.viewer.presentation()))
        self.viewer.presentation().setSearchFilter(self.task.subject())
        self.assertEqual(1, len(self.viewer.presentation()))
        self.viewer.presentation().setSearchFilter(self.task.subject(), 
                                                   includeSubItems=True)
        self.assertEqual(2, len(self.viewer.presentation()))
        
    def testAscendingSortOrder(self):
        self.task.addEffort(self.effort1)
        self.task.addEffort(self.effort2)
        self.viewer.setSortOrderAscending(True)
        self.assertEqual([self.effort1, self.effort2], list(self.viewer.presentation()))

    def testDescendingSortOrder(self):
        self.task.addEffort(self.effort1)
        self.task.addEffort(self.effort2)
        self.viewer.setSortOrderAscending(False)
        self.assertEqual([self.effort2, self.effort1], list(self.viewer.presentation()))
        
        
class EffortViewerAggregationTestCase(test.wxTestCase):
    aggregation = 'Subclass responsibility'
    
    def createViewer(self):
        return gui.viewer.EffortViewer(self.frame, self.taskFile, self.settings)

    def setUp(self):
        super(EffortViewerAggregationTestCase, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.settings.set('effortviewer', 'aggregation', self.aggregation)

        self.taskFile = persistence.TaskFile()
        self.viewer = self.createViewer()
        self.task = task.Task('Task')
        self.task2 = task.Task('Task2')
        self.taskFile.tasks().extend([self.task, self.task2])
        self.task.addEffort(effort.Effort(self.task, 
            date.DateTime(2008, 7, 16, 10, 0, 0), 
            date.DateTime(2008, 7, 16, 11, 0, 0)))
        self.task.addEffort(effort.Effort(self.task, 
            date.DateTime(2008, 7, 16, 12, 0, 0), 
            date.DateTime(2008, 7, 16, 13, 0, 0)))
        self.task.addEffort(effort.Effort(self.task,
            date.DateTime(2008, 7, 17, 1, 0, 0), 
            date.DateTime(2008, 7, 17, 2, 0, 0)))
        mostRecentPeriod = (date.DateTime(2008, 7, 23, 1, 0, 0), 
                            date.DateTime(2008, 7, 23, 2, 0, 0))
        # pylint: disable=W0142
        self.task.addEffort(effort.Effort(self.task, *mostRecentPeriod))
        self.task2.addEffort(effort.Effort(self.task2, *mostRecentPeriod))
        

    def tearDown(self):
        super(EffortViewerAggregationTestCase, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def switchAggregation(self):
        aggregations = ['details', 'day', 'week', 'month']
        aggregations.remove(self.aggregation)
        self.settings.settext(self.viewer.settingsSection(), 'aggregation', 
                              aggregations[0])


class EffortViewerAggregationRoundingTestCase(test.wxTestCase):
    aggregation = 'Subclass responsibility'
    roundingValue = None
    alwaysRoundUp = None

    def createViewer(self):
        return gui.viewer.EffortViewer(self.frame, self.taskFile, self.settings)

    def setUp(self):
        super(EffortViewerAggregationRoundingTestCase, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.settings.set('effortviewer', 'aggregation', self.aggregation)
        self.settings.setint('effortviewer', 'round', self.roundingValue)
        self.settings.setboolean('effortviewer', 'alwaysroundup', self.alwaysRoundUp)

        self.taskFile = persistence.TaskFile()
        self.viewer = self.createViewer()
        self.task = task.Task('Task')
        self.taskFile.tasks().extend([self.task])
        self.task.addEffort(effort.Effort(self.task,
                                          date.DateTime(2013, 7, 6, 1, 0, 0),
                                          date.DateTime(2013, 7, 6, 1, 0, 45)))
        self.task.addEffort(effort.Effort(self.task,
                                          date.DateTime(2013, 7, 6, 2, 0, 0),
                                          date.DateTime(2013, 7, 6, 2, 0, 45)))
        self.task.addEffort(effort.Effort(self.task,
                                          date.DateTime(2013, 7, 6, 3, 0, 0),
                                          date.DateTime(2013, 7, 6, 3, 0, 45)))
        self.task.addEffort(effort.Effort(self.task,
                                          date.DateTime(2013, 7, 6, 4, 0, 0),
                                          date.DateTime(2013, 7, 6, 4, 0, 10)))


class RoundingTestsMixin(object):
    def testRenderDuration(self):
        self.assertEqual(self.expectedPeriodRendering, self.viewer.widget.getItemText(self.viewer.widget.getItemWithIndex(0), 3))


class EffortViewerAggregationRoundingDayTest(EffortViewerAggregationRoundingTestCase, RoundingTestsMixin):
    aggregation = 'day'
    roundingValue = 60
    alwaysRoundUp = False
    expectedPeriodRendering = '0:03'


class EffortViewerAggregationRoundingDayUpTest(EffortViewerAggregationRoundingTestCase, RoundingTestsMixin):
    aggregation = 'day'
    roundingValue = 60
    alwaysRoundUp = True
    expectedPeriodRendering = '0:04'


class EffortViewerAggregationRoundingWeekTest(EffortViewerAggregationRoundingTestCase, RoundingTestsMixin):
    aggregation = 'week'
    roundingValue = 60
    alwaysRoundUp = False
    expectedPeriodRendering = '0:03'


class EffortViewerAggregationRoundingWeekUpTest(EffortViewerAggregationRoundingTestCase, RoundingTestsMixin):
    aggregation = 'week'
    roundingValue = 60
    alwaysRoundUp = True
    expectedPeriodRendering = '0:04'


class EffortViewerAggregationRoundingMonthTest(EffortViewerAggregationRoundingTestCase, RoundingTestsMixin):
    aggregation = 'month'
    roundingValue = 60
    alwaysRoundUp = False
    expectedPeriodRendering = '0:03'


class EffortViewerAggregationRoundingMonthUpTest(EffortViewerAggregationRoundingTestCase, RoundingTestsMixin):
    aggregation = 'month'
    roundingValue = 60
    alwaysRoundUp = True
    expectedPeriodRendering = '0:04'


class EffortViewerAggregationNoRoundingTest(EffortViewerAggregationRoundingTestCase, RoundingTestsMixin):
    aggregation = 'details'
    roundingValue = 60
    alwaysRoundUp = False
    expectedPeriodRendering = '0:00:10'


class EffortViewerAggregationNoRoundingUpTest(EffortViewerAggregationRoundingTestCase, RoundingTestsMixin):
    aggregation = 'details'
    roundingValue = 60
    alwaysRoundUp = True
    expectedPeriodRendering = '0:00:10'


class CommonTestsMixin(object):
    def testNumberOfItems(self):
        self.assertEqual(self.expectedNumberOfItems, self.viewer.size())

    def testRenderPeriod(self):
        self.assertEqual(self.expectedPeriodRendering, 
                         self.viewer.widget.GetItemText(0))

    def testRenderRepeatedPeriod(self):
        self.assertEqual('', self.viewer.widget.GetItemText(1))

    def testSwitchAggregation(self):
        self.switchAggregation()
        self.settings.settext(self.viewer.settingsSection(), 'aggregation', 
                              self.aggregation)
        self.assertEqual(self.expectedNumberOfItems, self.viewer.size())

    def testAggregationIsSavedInSettings(self):
        self.assertEqual(self.aggregation, 
            self.settings.get(self.viewer.settingsSection(), 'aggregation'))

    def testToolbarChoiceCtrlShowsAggegrationMode(self):
        aggregationUICommand = self.viewer.aggregationUICommand
        index = aggregationUICommand.choiceData.index(self.aggregation)
        expectedLabel = aggregationUICommand.choiceLabels[index]
        actualLabel = aggregationUICommand.choiceCtrl.GetStringSelection()
        self.assertEqual(expectedLabel, actualLabel)
        
    def testSearch(self):
        self.viewer.setSearchFilter('Task2')
        self.assertEqual(1, self.viewer.size())
        
    def testSearchDescription(self):
        self.task.efforts()[0].setDescription('Description')
        self.viewer.setSearchFilter('Description', searchDescription=True)
        self.assertEqual(1, self.viewer.size())

    def testSearchWithIncludeSubitems(self):
        self.viewer.setSearchFilter('Task2', includeSubItems=True)
        self.assertEqual(1, self.viewer.size())
        
    def testDelete(self):
        self.viewer.widget.Select(0)
        self.viewer.updateSelection()
        self.viewer.deleteUICommand.doCommand(None)
        expectedNumberOfItems = self.expectedNumberOfItems - (1 if self.aggregation == 'details' else 3)
        self.assertEqual(expectedNumberOfItems, self.viewer.size())
    
    def testDeleteTask(self):
        self.taskFile.tasks().remove(self.task2)
        expectedNumberOfItems = self.expectedNumberOfItems - 1
        self.assertEqual(expectedNumberOfItems, self.viewer.size())
        
    def testNewEffortUsesSameTaskAsSelectedEffort(self):
        dialog = self.viewer.newItemDialog(selectedTasks=[self.task2], 
                                           bitmap='new')
        for newEffort in dialog._items:  # pylint: disable=W0212
            self.assertEqual(self.task2, newEffort.task())
        
    def testColumnUICommands(self):
        expectedLength = dict(details=6, day=8, week=9, month=8)[self.aggregation]
        self.assertEqual(expectedLength,
                         len(self.viewer.getColumnUICommands()))
        
    def testTotalTimeSpentColumnNotInDetailsMode(self):
        columns = [command.setting for command in self.viewer.getColumnUICommands() if command]
        self.assertEqual(self.aggregation != 'details', 
                         'totalTimeSpent' in columns)

    def testTotalRevenueColumnNotInDetailsMode(self):
        columns = [command.setting for command in self.viewer.getColumnUICommands() if command]
        self.assertEqual(self.aggregation != 'details', 
                         'totalRevenue' in columns)
    
    def testDefaultNrOfColumns(self):
        self.assertEqual(4, self.viewer.widget.GetColumnCount())

    def testHideTimeSpentColumn(self):
        self.viewer.showColumnByName('timeSpent', False)
        self.assertEqual(3, self.viewer.widget.GetColumnCount())
        
    def testHideRevenueColumn(self):
        self.viewer.showColumnByName('revenue', False)
        self.assertEqual(4, self.viewer.widget.GetColumnCount())
        
    def testShowTotalTimeSpentColumn(self):
        self.viewer.showColumnByName('totalTimeSpent', True)
        self.assertEqual(5, self.viewer.widget.GetColumnCount())

    def testShowTotalRevenueColumn(self):
        self.viewer.showColumnByName('totalRevenue', True)
        self.assertEqual(5, self.viewer.widget.GetColumnCount())
        
    def testTotalTimeSpentColumnIsHiddenWhenSwitchingToDetails(self):
        self.viewer.showColumnByName('totalTimeSpent', True)
        self.switchAggregation()
        self.assertEqual(self.viewer.isShowingAggregatedEffort(),
                         self.viewer.isVisibleColumnByName('totalTimeSpent'))

    def testTotalRevenueColumnIsHiddenWhenSwitchingToDetails(self):
        self.viewer.showColumnByName('totalRevenue', True)
        self.switchAggregation()
        self.assertEqual(self.viewer.isShowingAggregatedEffort(),
                         self.viewer.isVisibleColumnByName('totalRevenue'))

    def testActiveEffort(self):
        self.task2.efforts()[0].setStop(date.DateTime.max)  # Make active
        self.viewer.secondRefresher.onEverySecond()  # Simulate clock firing
        expectedNrOfTrackedItems = 1 if self.aggregation == 'details' else 2
        self.assertEqual(expectedNrOfTrackedItems, 
                         len(self.viewer.secondRefresher.currentlyTrackedItems()))
        
    def testActiveEffortAfterSwitch(self):
        self.task2.efforts()[0].setStop(date.DateTime.max)  # Make active
        self.switchAggregation()    
        self.viewer.secondRefresher.onEverySecond()  # Simulate clock firing
        expectedNrOfTrackedItems = 2 if self.aggregation == 'details' else 1
        self.assertEqual(expectedNrOfTrackedItems, 
                         len(self.viewer.secondRefresher.currentlyTrackedItems()))
        
    def testIsShowingAggregatedEffort(self):
        isAggregating = self.aggregation != 'details'
        self.assertEqual(isAggregating, self.viewer.isShowingAggregatedEffort())
        
    def testStopEffortTracking(self):
        self.task.addEffort(effort.Effort(self.task))
        stopUICommand = gui.uicommand.EffortStop(viewer=self.viewer,
                                                 effortList=self.taskFile.efforts(),
                                                 taskList=self.taskFile.tasks())
        stopUICommand.doCommand()
        self.failIf(self.task.isBeingTracked())
    

class EffortViewerWithoutAggregationTest(CommonTestsMixin, 
                                         EffortViewerAggregationTestCase):
    aggregation = 'details'
    expectedNumberOfItems = 5
    expectedPeriodRendering = render.dateTimePeriod(\
        date.DateTime(2008, 7, 23, 1, 0), date.DateTime(2008, 7, 23, 2, 0))
    
    
class EffortViewerWithAggregationPerDayTest(CommonTestsMixin, 
                                            EffortViewerAggregationTestCase):
    aggregation = 'day'
    expectedNumberOfItems = 7  # 4 day/task combinations on 3 days (== 3 total rows) 
    expectedPeriodRendering = render.date(date.DateTime(2008, 7, 23))


class EffortViewerWithAggregationPerWeekTest(CommonTestsMixin, 
                                             EffortViewerAggregationTestCase):
    aggregation = 'week'
    expectedNumberOfItems = 5  # 3 week/task combinations in 2 weeks (== 2 total rows)
    expectedPeriodRendering = '2008-30'


class EffortViewerWithAggregationPerMonthTest(CommonTestsMixin, 
                                              EffortViewerAggregationTestCase):
    aggregation = 'month'
    expectedNumberOfItems = 3  # 2 month/task combinations in 1 month (== 1 total row)
    expectedPeriodRendering = render.month(date.DateTime(2008, 07, 01))


class EffortViewerRenderTestMixin(object):
    aggregation = 'Subclass responsibility'

    def createViewer(self):
        return gui.viewer.EffortViewer(self.frame, self.taskFile, self.settings)

    def setUp(self):
        super(EffortViewerRenderTestMixin, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.settings.set('effortviewer', 'aggregation', self.aggregation)

        self.taskFile = persistence.TaskFile()
        self.task = task.Task('task')
        self.taskFile.tasks().append(self.task)
        self.midnight = date.Now().startOfDay()
        self.viewer = self.createViewer()

    def testToday(self):
        theEffort = effort.Effort(self.task, self.midnight, self.midnight + date.TWO_HOURS)
        self.task.addEffort(theEffort)
        text = self.viewer.widget.GetItemText(0)
        self.failUnless(text.startswith('Today'), '"Today" not in %s' % text)

    def testTomorrow(self):
        theEffort = effort.Effort(self.task, self.midnight + date.ONE_DAY,
                                  self.midnight + date.TimeDelta(hours=2, days=1))
        self.task.addEffort(theEffort)
        text = self.viewer.widget.GetItemText(0)
        self.failUnless(text.startswith('Tomorrow'), '"Tomorrow" not in %s' % text)

    def testYesterday(self):
        theEffort = effort.Effort(self.task, self.midnight - date.TimeDelta(days=1),
                                  self.midnight - date.TimeDelta(hours=22))
        self.task.addEffort(theEffort)
        text = self.viewer.widget.GetItemText(0)
        self.failUnless(text.startswith('Yesterday'), '"Yesterday" not in %s' % text)


class EffortViewerRenderDetailsTest(EffortViewerRenderTestMixin, test.wxTestCase):
    aggregation = 'details'


class EffortViewerRenderPerDayTest(EffortViewerRenderTestMixin, test.wxTestCase):
    aggregation = 'day'
